 // ################################################################################
//
//  Sumo Robot SX v0.0
//
//  Released:  22/12/2020
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    microcontroller:  ESP32 Dev Kit

    This code controls a miniature Sumo robot, which has a range of programmeable
    features, which inhcludes the simultaneous use of the following function:
    * 128x64 OLED display
    * LTOF laser range sensor
    * MPU6050 Gyro sensor
    * Neopixel RGB LEDs
    * Infrared line sensors
    * micro servo PWM
    * LX-16A serial servo
    * ESP-NOW WiFi link to Wii Transceiver
    * button switches
    * EEPROM stored reference data
    * AI task scheduler
    * Move engine

    Conditional I/O compilation based on SumoBot number.

*/
// Declare Libraries
#include <Arduino.h>
#include <HardwareSerial.h>     // serial library
#include <Wire.h>               // I2C library
#include "SSD1306Wire.h"        // OLED display library
#include <Adafruit_NeoPixel.h>  // Neopixel library
#include <ESP32Servo.h>         // servo library
#include <esp_now.h>            // ESP-NOW library
#include <WiFi.h>               // ESP WiFi library
#include <EEPROM.h>             // EEPROM flash access library

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = 21;
const int SCL_PIN = 22;

// Initialize the oled display for address 0x3c
SSD1306Wire display(I2C_DISPLAY_ADDRESS, SDA_PIN, SCL_PIN);

// define VL53LOX constants
#define ADDRESS_DEFAULT 0b0101001 // I2C address 0x29 HEX
#define calcMacroPeriod(vcsel_period_pclks) ((((uint32_t)2304 * (vcsel_period_pclks)* 1655) + 500) / 1000)
#define checkTimeoutExpired() (io_timeout > 0 && ((uint16_t)millis() - timeout_start_ms) > io_timeout)
#define decodeVcselPeriod(reg_val) (((reg_val) + 1) << 1)
#define encodeVcselPeriod(period_pclks) (((period_pclks) >> 1) - 1)
#define startTimeout() (timeout_start_ms = millis())
#define XSHUT 18  // GPIO18 is connected to the XSHUT pin

// Declare register list from VL53L0X.h API, ordered as listed there
#define SYSRANGE_START 0x00
#define SYSTEM_THRESH_HIGH 0x0C
#define SYSTEM_THRESH_LOW 0x0E
#define SYSTEM_SEQUENCE_CONFIG 0x01
#define SYSTEM_RANGE_CONFIG 0x09
#define SYSTEM_INTERMEASUREMENT_PERIOD 0x04
#define SYSTEM_INTERRUPT_CONFIG_GPIO 0x0A
#define GPIO_HV_MUX_ACTIVE_HIGH 0x84
#define SYSTEM_INTERRUPT_CLEAR 0x0B
#define RESULT_INTERRUPT_STATUS 0x13
#define RESULT_RANGE_STATUS 0x14
#define RESULT_CORE_AMBIENT_WINDOW_EVENTS_RTN 0xBC
#define RESULT_CORE_RANGING_TOTAL_EVENTS_RTN 0xC0
#define RESULT_CORE_AMBIENT_WINDOW_EVENTS_REF 0xD0
#define RESULT_CORE_RANGING_TOTAL_EVENTS_REF 0xD4
#define RESULT_PEAK_SIGNAL_RATE_REF 0xB6
#define ALGO_PART_TO_PART_RANGE_OFFSET_MM 0x28
#define I2C_SLAVE_DEVICE_ADDRESS 0x8A
#define MSRC_CONFIG_CONTROL 0x60
#define PRE_RANGE_CONFIG_MIN_SNR 0x27
#define PRE_RANGE_CONFIG_VALID_PHASE_LOW 0x56
#define PRE_RANGE_CONFIG_VALID_PHASE_HIGH 0x57
#define PRE_RANGE_MIN_COUNT_RATE_RTN_LIMIT 0x64
#define FINAL_RANGE_CONFIG_MIN_SNR 0x67
#define FINAL_RANGE_CONFIG_VALID_PHASE_LOW 0x47
#define FINAL_RANGE_CONFIG_VALID_PHASE_HIGH 0x48
#define FINAL_RANGE_CONFIG_MIN_COUNT_RATE_RTN_LIMIT 0x44
#define PRE_RANGE_CONFIG_SIGMA_THRESH_HI 0x61
#define PRE_RANGE_CONFIG_SIGMA_THRESH_LO 0x62
#define PRE_RANGE_CONFIG_VCSEL_PERIOD 0x50
#define PRE_RANGE_CONFIG_TIMEOUT_MACROP_HI 0x51
#define PRE_RANGE_CONFIG_TIMEOUT_MACROP_LO 0x52
#define SYSTEM_HISTOGRAM_BIN 0x81
#define HISTOGRAM_CONFIG_INITIAL_PHASE_SELECT 0x33
#define HISTOGRAM_CONFIG_READOUT_CTRL 0x55
#define FINAL_RANGE_CONFIG_VCSEL_PERIOD 0x70
#define FINAL_RANGE_CONFIG_TIMEOUT_MACROP_HI 0x71
#define FINAL_RANGE_CONFIG_TIMEOUT_MACROP_LO 0x72
#define CROSSTALK_COMPENSATION_PEAK_RATE_MCPS 0x20
#define MSRC_CONFIG_TIMEOUT_MACROP 0x46
#define SOFT_RESET_GO2_SOFT_RESET_N 0xBF
#define IDENTIFICATION_MODEL_ID 0xC0
#define IDENTIFICATION_REVISION_ID 0xC2
#define OSC_CALIBRATE_VAL 0xF8
#define GLOBAL_CONFIG_VCSEL_WIDTH 0x32
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_0 0xB0
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_1 0xB1
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_2 0xB2
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_3 0xB3
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_4 0xB4
#define GLOBAL_CONFIG_SPAD_ENABLES_REF_5 0xB5
#define GLOBAL_CONFIG_REF_EN_START_SELECT 0xB6
#define DYNAMIC_SPAD_NUM_REQUESTED_REF_SPAD 0x4E
#define DYNAMIC_SPAD_REF_EN_START_OFFSET 0x4F
#define POWER_MANAGEMENT_GO1_POWER_FORCE 0x80
#define VHV_CONFIG_PAD_SCL_SDA__EXTSUP_HV 0x89
#define ALGO_PHASECAL_LIM 0x30
#define ALGO_PHASECAL_CONFIG_TIMEOUT 0x30

// define VL53L0C data structures
struct SequenceStepEnables {boolean tcc, msrc, dss, pre_range, final_range;};
struct SequenceStepTimeouts {
      uint16_t pre_range_vcsel_period_pclks, final_range_vcsel_period_pclks;
      uint16_t msrc_dss_tcc_mclks, pre_range_mclks, final_range_mclks;
      uint32_t msrc_dss_tcc_us, pre_range_us, final_range_us;};

// declare VL53L0X variables
uint8_t address = ADDRESS_DEFAULT;  // this can be changed using VL53L0X_setAddress(uint8_t new_addr)
bool did_timeout;
uint16_t io_timeout;
uint8_t last_status;      // status of last I2C transmission
unsigned long LTOF_t0;    // fps timer
unsigned long LTOF_t1;    // fps timer
int16_t LTOF_fps;         // LTOF fps value
int16_t LTOF_On;          // > 0 when laser ranging is enabled
unsigned long LTOF_Wait;  // delay used to prevent wasted time in LTOF readings
uint32_t measurement_timing_budget_us;
byte Pulse = 0;           // unfiltered Range transient pulse counter
uint16_t Range;           // raw 16-bit range value
int RangeFtd;             // filtered range value used by general code
int RangeFtdMin;          // temporary minimum filtered range, used to find minimum
bool RangeFtdNew;         // set == true when a new filtered range is determined
int RangeFtdTgt;          // range to detected min target
int RangeFtdTgt10;        // detected min range + 10mm
uint16_t RangeLast;       // 16-bit range value
bool RangeNew;            // set to true when a new Range value is determined
uint8_t Status;           // range status data, indicates quality of range value
uint8_t stop_variable;    // read by init and used when starting measurement; is StopVariable field of VL53L0X_DevData_t structure in API
uint16_t timeout_start_ms;
enum vcselPeriodType {VcselPeriodPreRange, VcselPeriodFinalRange};
void VL53L0X_startContinuous(uint32_t period_ms = 0);
inline void VL53L0X_setTimeout(uint16_t timeout) { io_timeout = timeout; }
bool VL53L0X_timeoutOccurred(void);
bool VL53L0X_getSpadInfo(uint8_t * count, bool * type_is_aperture);


// 3-axis MPU 6050 variables
const int MPU_addr=0x68;      // I2C address of the MPU-6050
long AcAvgDiv;                // dividing factor used in averaging
long AcAvgDivMax;             // max dividing factor for averaging
int16_t AcX;                  // MPU raw X-acceleration value
long AcXAvg;                  // averaged AcX value to reduce noise
int16_t AcXL;                 // previous MPU X-acceleration value
int16_t AcXOff;               // offset to be added to raw X-acceleration value
long AcXSum;                  // averaging accumulator
int16_t AcY;                  // MPU raw Y-acceleration value
long AcYAvg;                  // averaged AcY value to reduce noise
float AcYF;                   // floating point Y acceleration value
int16_t AcYL;                 // previous MPU Y-acceleration value
int16_t AcYOff;               // offset to be added to raw X-acceleration value
bool AcYsgn;                  // positive/negative sign flag
long AcYSum;                  // averaging accumulator
int16_t AcZ;                  // MPU raw Z-acceleration value
long AcZAvg;                  // averaged AcZ value to reduce noise
int16_t AcZL;                 // previous MPU Z-acceleration value
int16_t AcZOff;               // offset to be added to raw X-acceleration value
long AcZSum;                  // averaging accumulator
long GyAvgDiv;                // dividing factor used in averaging
int GyCal;                    // counter used to determine gyro calibration period
unsigned long GyT;            // time of gyro readings in milliseconds
unsigned long GyTD;           // time between successive gyro readings in milliseconds
int16_t GyX;                  // MPU raw X-gyroscope value
long GyXAvg;                  // averaged GyX value to reduce noise
int16_t GyXL;                 // previous MPU X-gyroscope value
int16_t GyXOff;               // offset to be added to raw X-gyroscope value
int16_t GyXOffCnt;            // offset counter for gyro centering
long GyXSum;                  // averaging accumulator
int16_t GyY;                  // MPU raw Y-gyroscope value
long GyYAvg;                  // averaged GyY value to reduce noise
int16_t GyYL;                 // previous MPU Y-gyroscope value
int16_t GyYOff;               // offset to be added to raw Y-gyroscope value
int16_t GyYOffCnt;            // offset counter for gyro centering
long GyYSum;                  // averaging accumulator
int16_t GyZ;                  // MPU raw Z-gyroscope value
long GyZAvg;                  // averaged GyZ value to reduce noise
int16_t GyZL;                 // previous MPU Z-gyroscope value
int16_t GyZOff;               // offset to be added to raw Z-gyroscope value
int16_t GyZOffCnt;            // offset counter for gyro centering
long GyZSum;                  // averaging accumulator
bool MPUCal;                  // == true during calibration
float PitchAcc;               // pitch angle in degrees base on accelerometer
float PitchGyr;               // pitch angle in degrees base on gyro + acc
int PitchGyrInt;              // integer version of PitchGyr for quicker calcs
bool UpRht;                   // set == true when the robot is the correct way up
float YawAcc;                 // yaw angle in degrees based on accelerometers
float YawGyr;                 // yaw angle in degrees based on gyro + acc
int YawGyrInt;                // integer version of YawGyr for quicker calcs

// declare general constants
#define BatCal 377.1          // calibrated voltage multiplying factor, default 404.85 
#define BatCritical 2489      // critical battery threshold @6.6v, default = 2672, set == 0 to ignore
#define BatMax 3092           // A0 for fully charged battery voltage, == 8.2v
#define BatPin 33             // analog input pin ADC5 on GPIO33
#define BatUSB 1858           // if <= then USB supply assumed
#define EEPROM_SIZE 38        // define the number of bytes you want to access
#define GyAvgDivMax 250       // max averaging divider, 250 == 10 second average
#define IR_Black 2000         // min sensor level for Black
#define IR_White 150          // max sensor level for white
#define IR_White5 750         // 5 x IR_White
#define NeoPin 19             // assign GPIO19 for RGB LEDs
#define RangeMax 600          // maximum range 'filtered'
#define RangeMin 50           // minimum range 'filtered'
#define RangeTgtAtt 150       // target range which triggers and attack
#define RangeTgtOR 500        // upper limit of range targeting, or out of range
#define RangeTgtUL 350        // limit of detectable target in mm
#define RangeOL 1200          // range outer limit, 1200 mm 
#define ServoAngMaxDef 130    // micro-servo max angle
#define ServoAngMinDef 50     // micro-servo min angle
#define ServoAngWidth 80      // micro-servo angle range
#define ServoCtrDef 1500      // micro servo PWM centre cal. limit
#define ServoMinDef 900       // micro servo PWM lower cal. limit
#define ServoMaxDef 2100      // micro servo PWM upper cal. limit
#define servoPin 4            // allocate servo pin as GPIO4
#define SpeedMax 1000         // max LX-16A speed
#define SpeedMid 400          // mid LX-16A speed
#define SpeedMin 150          // min LX-16A speed
#define sw0Pin 13             // left switch SW0 assigned to GPIO13
#define sw1Pin 15             // right switch SW1 assigned to GPIO15

#define SumoBot 1             // change this to suit your robot wiring
#if SumoBot == 1
  // Sumo Bot 1
  // declare IR sensor pin and channel mappings
  #define FLftPwr 14  // IR sensor front left power on GPIO14
  #define FLftADC 32  // IR sensor front left A0 on ADC4
  #define FLftACh A4  // ADC channel assigned to front left sensor
  #define FCtrPwr 27  // IR sensor front centre power on GPIO27
  #define FCtrADC 35  // IR sensor front centre A0 on ADC7
  #define FCtrACh A7  // ADC channel assigned to front centre sensor
  #define FRhtPwr 26  // IR sensor front right power on GPIO26
  #define FRhtADC 39  // IR sensor front left A0 on ADC3
  #define FRhtACh A3  // ADC channel assigned to front right sensor
  #define RRhtPwr 25  // IR sensor rear right power on GPIO25
  #define RRhtADC 34  // IR sensor rear right A0 on ADC6
  #define RRhtACh A6  // ADC channel assigned to rear right sensor
  #define RLftPwr 12  // IR sensor rear left power on GPIO12
  #define RLftADC 36  // IR sensor rear left A0 on ADC0
  #define RLftACh A0  // ADC channel assigned to rear right sensor
#endif

#if SumoBot == 2
  //  Sumo Bot 2
  //  declare IR sensor pin and channel mappings
  #define FLftPwr 27  // IR sensor front left power on GPIO14
  #define FLftADC 34  // IR sensor front left A0 on ADC4
  #define FLftACh A6  // ADC channel assigned to front left sensor
  #define FCtrPwr 12  // IR sensor front centre power on GPIO27
  #define FCtrADC 32  // IR sensor front centre A0 on ADC7
  #define FCtrACh A4  // ADC channel assigned to front centre sensor
  #define FRhtPwr 25  // IR sensor front right power on GPIO26
  #define FRhtADC 36  // IR sensor front left A0 on ADC3
  #define FRhtACh A0  // ADC channel assigned to front right sensor
  #define RRhtPwr 14  // IR sensor rear right power on GPIO25
  #define RRhtADC 35  // IR sensor rear right A0 on ADC6
  #define RRhtACh A7  // ADC channel assigned to rear right sensor
  #define RLftPwr 26  // IR sensor rear left power on GPIO12
  #define RLftADC 39  // IR sensor rear left A0 on ADC0
  #define RLftACh A3  // ADC channel assigned to rear right sensor
#endif
 
// When we setup the NeoPixel library, we tell it how many pixels, and which pin to use to send signals.
// Note that for older NeoPixel strips you might need to change the third parameter--see the strandtest
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(5, NeoPin, NEO_GRB + NEO_KHZ800);

// create servo instance
Servo microServo;  // create servo object to control a servo

// LX-16A functions
#define GET_LOW_BYTE(A) (uint8_t)((A))        //Macro function  get lower 8 bits of A
#define GET_HIGH_BYTE(A) (uint8_t)((A) >> 8)  //Macro function  get higher 8 bits of A
#define BYTE_TO_HW(A, B) ((((uint16_t)(A)) << 8) | (uint8_t)(B))  //put A as higher 8 bits   B as lower 8 bits   which amalgamated into 16 bits integer

// LX-16A command constants
#define LOBOT_SERVO_FRAME_HEADER         0x55
#define LOBOT_SERVO_MOVE_TIME_WRITE      1
#define LOBOT_SERVO_MOVE_TIME_READ       2
#define LOBOT_SERVO_MOVE_TIME_WAIT_WRITE 7
#define LOBOT_SERVO_MOVE_TIME_WAIT_READ  8
#define LOBOT_SERVO_MOVE_START           11
#define LOBOT_SERVO_MOVE_STOP            12
#define LOBOT_SERVO_ID_WRITE             13
#define LOBOT_SERVO_ID_READ              14
#define LOBOT_SERVO_ANGLE_OFFSET_ADJUST  17
#define LOBOT_SERVO_ANGLE_OFFSET_WRITE   18
#define LOBOT_SERVO_ANGLE_OFFSET_READ    19
#define LOBOT_SERVO_ANGLE_LIMIT_WRITE    20
#define LOBOT_SERVO_ANGLE_LIMIT_READ     21
#define LOBOT_SERVO_VIN_LIMIT_WRITE      22
#define LOBOT_SERVO_VIN_LIMIT_READ       23
#define LOBOT_SERVO_TEMP_MAX_LIMIT_WRITE 24
#define LOBOT_SERVO_TEMP_MAX_LIMIT_READ  25
#define LOBOT_SERVO_TEMP_READ            26
#define LOBOT_SERVO_VIN_READ             27
#define LOBOT_SERVO_POS_READ             28
#define LOBOT_SERVO_OR_MOTOR_MODE_WRITE  29
#define LOBOT_SERVO_OR_MOTOR_MODE_READ   30
#define LOBOT_SERVO_LOAD_OR_UNLOAD_WRITE 31
#define LOBOT_SERVO_LOAD_OR_UNLOAD_READ  32
#define LOBOT_SERVO_LED_CTRL_WRITE       33
#define LOBOT_SERVO_LED_CTRL_READ        34
#define LOBOT_SERVO_LED_ERROR_WRITE      35
#define LOBOT_SERVO_LED_ERROR_READ       36

// LX16-A code constants
#define ID1   1 // servo reference for left-hand servo
#define ID2   2 // servo reference for right-hand servo
//#define LOBOT_DEBUG 1  // Debug ：print debug value

// Wii Transciever MAC: 50:02:91:68:F7:3F
uint8_t broadcastAddress[] = {0x50, 0x02, 0x91, 0x68, 0xF7, 0x3F};

// 250 byte max for Tx/Rx data
// It must match the receivers data structure
typedef struct ESPNOW_Buff250 {
    char ESPdata[250];
} ESPNOW_Buff250;

// Create an 8Byte_message called NunchukDataTx to hold I2C data message
ESPNOW_Buff250 Tx_Buff;

// Create an 8Byte_message to hold incoming sensor readings
ESPNOW_Buff250 Rx_Buff;

// Declare and initialise global variables
int16_t ADC;                  // IR sensor pointer, 0 - 4
int16_t ADC_Cnt;              // timer counter used in IR sensor reading
int16_t ADC_Data[40];         // sensor data recorder
bool ADC_Fast;                // IR sensor read flag; fast == true
byte ADC_Rec;                 // ADC sensor channel to be recorded 0 - 4
int16_t ADC_RecPnt;           // ADC data array pointer used in storing data
bool ADCscan;                 // true once a complete scan has been performed
bool AI_ATTACK_EN;            // if == true it enables AI attack mode
int16_t AI_ATT_CNT;           // sustrains the attacking process
bool AI_EDGE_ON;              // if == true AI performs edge detection
bool AI_IR_ON;                // if == true AI turns ON IR sensor scanning
int16_t AI_LED_Task;          // default AI LED task when in IR mode
bool AI_LTOF_ON;              // == true enables LTOF servo scanner in AI mode
bool AI_REPORT;               // if == true then report messages
int16_t AI_Sub;               // initial AI sub-task
bool AI_TIP_ON;               // == true enables AI tip responses
String AI_Title$;             // title used in AI display logo
int16_t AI_TurnCnt;           // timer counter used in AI steering
int16_t Any;                  // any temp value
long AnyLong;                 // any temp value
String Any$;                  // temporary string
bool Attack;                  // AI task attacking mode
int16_t Av100[] = {0,0,0,0,0};// sensor percentage levels
int16_t AvBlk[] = {IR_Black,IR_Black,IR_Black,IR_Black,IR_Black}; // sensor black thresholds
int16_t AvRaw[5];             // ADC readings for IR sensors
long AvTmp[5];                // temp values
int16_t AvWht[] = {IR_White,IR_White,IR_White,IR_White,IR_White}; // sensor white thresholds
int16_t BatAvg;               // preload averaging function with high value
int16_t BatAvgMax;            // maximum average value in this period
byte BatCnt;                  // pointer to array values and critical timer
int16_t BatCritCnt;           // timer used in baterycritical testing
int16_t BatData[60];          // battery averages at 1 second intervals
int16_t BatDel;               // set a 10 second delay in 'life' prediction
int16_t BatDP;                // battery data pointer, 0 - 59
int16_t BatLast;              // previous battery average reading
bool BattNow;                 // == true causes battery reading
bool BatPwr;                  // true if battery supply detected, false if USB assumed
int16_t BatQ;                 // a measure of battery quality 1 - 5
int16_t BatRead;              // timer counter used in reading battery
unsigned long BatRep;         // time when to report battery reading over serial link
long BatSum;                  // preload averaging sum with 20x max value
unsigned long BatT0;          // time of first battery measurement in life prediction forecast
String BatTime$;              // time left displayed
unsigned long BatTms;         // time when battery reading was taken
int16_t BatV0;                // reference voltage used in prediction forecast
int16_t BatV1;                // lowest voltage used in prediction forecast
int16_t BatVol;               // instantaneous battery voltage
int16_t C_Cnt;                // counter used in 'C' button detection
byte C_Dn;                    // counter for 'C' button down period
int16_t CountDown;            // seconds count down used by AI
byte C_Up;                    // counter for 'C' button up period
char cmdMode;                 // command mode
int16_t cmdSgn;               // cmdVal sign, 0 or -1
char cmdType;                 // command mode type
int16_t cmdVal;               // value associated with a cmdType
uint8_t colBlu[] = {0,0,0,0,0,0,0,0,0,0}; // mouth colour array - blue
uint8_t colB;                 // mouth colour flag - blue
uint8_t colGrn[] = {0,0,0,0,0,0,0,0,0,0}; // mouth colour array - green
uint8_t colG;                 // mouth colour flag - green
uint8_t colRed[] = {0,0,0,0,0,0,0,0,0,0}; // mouth colour array - red
uint8_t colR;                 // mouth colour flag - red
int16_t CZ;                   // received button values
bool CZ_Wait;                 // flag used for button release wait function
bool DispBack;                // flag used to draw a back arrow on the display
bool DispChng;                // flag invokes initialisation code
bool DispClr;                 // set == true to clear the memory contents
int16_t DispCnt;              // counter used in display updating
int16_t DispDel;              // >0 display delay counter, stops updates over I2C
bool DispDisp;                // set = true to immediately display memory contents
byte DispMenu;                // group type, 0 = sensors, 1 = tasks, etc
int16_t DispMode;             // the type of display for a given mode
bool DispNOP;                 // set == true to stop display I2C updates
int16_t DispNow;              // if > 0 display special modes
int16_t DispPnt;              // points at the current display mode
byte DispCX,DispCY;           // display cursor coordinates
unsigned long DistanceMs;     // millisecond timer for distance travelled
bool Drive_En;                // == true if random steering applies in AI tasks
bool Echo;                    // if true send print messages
String Error$ = "";           // any error message
bool ESC0;                    // SW0 ESC exit current Main Task flag
bool ESC1;                    // SW1 ESC exit current Main Task flag
byte I2CError;                // used to detect I2C transmission errors
bool IR_BaseF;                // base colour, true == white, false == black
byte IR_Mask;                 // an 8-bit mask which identifies edge detection
byte IR_Mask2;                // a 2-bit mask which identifies rear edge detection
byte IR_Mask3;                // a 3-bit mask which identifies front edge detection
bool ItsOver;                 // == true if the bot has fallen over
int16_t JoyD;                 // direction value 0 - 8
int16_t JoyDL;                // direction 1 counter
int16_t JoyV;                 // joystick vector  = SQRT(SQ(JX) + SQ(JY))
int16_t JoyX;                 // received value
int16_t JoyXC;                // received value offset
int16_t JoyY;                 // received value
int16_t JoyYC;                // received value offset
char keyChar;                 // any keyboard character
int16_t keyVal;               // any keyboard value
long L_Dst;                   // distance travelled in 0.0001mm increments
bool LED_Blink;               // flag toggled to create blinking effects
int16_t LED_Cnt;              // general counter used in LED tasks
int16_t LED_Del;              // LED Task delay, adds pauses into the processes
int16_t LED_Inc;              // general increment used in LED tasks
int16_t LED_Last;             // previous LED task pointer
byte LED_IR_Mask;             // previous IR_Mask used to block unnecessary updates
bool LED_ON;                  // state of LED pin HIGH/LOW
bool LED_Nop;                 // == true to block LED tasks, default = false
int16_t LED_Pnt;              // general LED pointer
bool LEDshow;                 // LED show flag for mouth LEDs
int16_t LED_SubTask;          // LED sub-task pointer
bool LED_SwDwn;               // == true if a button switch has been pressed
int16_t LED_Task;             // LED task pointer
byte LEDVal;                  // values used in LED tasks
int16_t L_Spd;                // speed value sent to left LX-16A servo
int16_t L_SpdTgt;             // target speed for left LX-16A servo
int16_t LX16Phase;            // either 0 or 1 for LX16 up/down speed
int16_t LX16Spd;              // speed value sent tot he servo, +/- 0 - 1000
int16_t LX16Tgt;              // servo target number for data transfer
int16_t MainCnt;              // Main Task general counter/timer
int16_t MainDel;              // Main task delay counter, if > 0 holds off tasks
int16_t MainFor;              // Main task for loop counter
int16_t MainInt;              // Main task temp integer
int16_t MainMax,MainMin;      // any min/max temp values
int16_t MainNext;             // next MainTask to be performed after a delay
int16_t MainSub;              // Main Task sub-task
int16_t MainSum;              // general accumulator for main tasks
int16_t MainTask;             // main task pointer
unsigned long MainTime;       // temporary millisecond timer used in tasks
int16_t MoveState;            // move state machine pointer
int16_t MoveSub;              // move sub task pointer
int16_t MoveTask;             // special move task pointer
String myMAC;                 // MAC address if this device
unsigned long next5ms;        //  5ms loop timer
unsigned long next10ms;       // 10ms loop timer
unsigned long next20ms;       // 20ms loop timer
unsigned long next40ms;       // 40ms loop timer
unsigned long next200ms;      // 200ms loop timer
unsigned long nextSec;        // 1 sec loop timer
int16_t overRun;              // counter used to control undemanded movement
int16_t Ping = 0;             // Rx 'ping' counter
unsigned long print40ms;      // 40ms variable print timer
int16_t PrintTgt;             // 0 == serial port, 1 == WiFi ESP-NOW
String PrintTx = "";          // printed strings
int16_t RangeData[1 + ServoAngWidth]; // range data associated with servo scanning
long R_Dst;                   // distance travelled in 0.0001mm increments
bool readWiiCall;             // flag prevents call stack overflow
bool Reboot;                  // set true when reboot is in progress to stop all tasks
byte ResetCnt;                // timer used to invoke soft RESET
long RndAhead;                // if > 0 then go straight ahead for RndAhead mm
int16_t RndDel;               // random driving task skip counter/timer
int16_t RndInc;               // speed increments used in random steering
int16_t RndTask;              // tasks performed by random driving
int16_t RndTurn_Flg;          // flag > 0 if neutral turning
int16_t RotCnt;               // auto zero delay counter for Z-axis
float RotDif;                 // rotation angle difference in degrees based on gyro + acc
bool RotESC;                  // rotation exit flag during edge detection
float RotGyr;                 // rotation angle in degrees based on gyro + acc
int16_t RotGyrInt;            // rotation angle in degrees as an integer
int16_t RotTgt;               // target angle in degrees, when turning
int16_t R_Spd;                // speed value sent to right LX-16A servo
int16_t R_SpdTgt;             // target speed for right LX-16A servo
byte RunCnt;                  // temp counter used in tasks
int16_t Rx_Chk;               // Rx checksum
int16_t RxCZ;                 // buffered received CZ value
int16_t RxJoyX;               // buffered received JoyX value
int16_t RxJoyY;               // buffered received JoyY value
int16_t Rx_len;               // length of WiFi received data block
int16_t Rx_Pnt;               // data pointer user by Rx state machine
bool RxRec;                   // set true when a valid frame of data is received
int16_t RxState;              // receiver state machine state
int16_t Rx_Task;              // task pointer user by Rx state machine
int16_t RxVal;                // value received from serial Rx
byte RxWiFi[6];               // array holding 6 bytes of Wii received data
int16_t ServoAng;             // micro-servo angle; default = 90
int16_t ServoAngMax;          // micro-servo max angle; default = 130
int16_t ServoAngMin;          // micro-servo max angle; default = 50
int16_t ServoAngTgt;          // servo angle of minimum target
int16_t ServoAngTgt10;        // temporary angle used in target tracking
int16_t ServoAngTgtAvg;       // averaged target position
int16_t ServoAngTgtLL;        // servo angle of minimum target lower limt
int16_t ServoAngTgtUL;        // servo angle of minimum target upper limt
int16_t ServoAngTmp;          // temporary servo angle of minimum target
bool ServoAuto;               // if false switch OFF servo auto-adjustment features
int16_t ServoCtr;             // centre Cal. PWM value for microservo
bool ServoEn;                 // == true when servo is attached
int16_t ServoInc;             // angular increment used in servo scanning
int16_t ServoLL;              // micro servo PWM lower limit
int16_t ServoMax;             // maximum Cal. PWM value for microservo
int16_t ServoMaxNew;          // maximum Cal. PWM value for microservo during Cal.
int16_t ServoMin;             // minimum Cal. PWM value for microservo
int16_t ServoMinNew;          // minimum Cal. PWM value for microservo during Cal.
int16_t servoPhase;           // either 0 or 1 for up/down angle change
int16_t ServoSubTask;         // servo sub-task pointer, default = 0
int16_t ServoTaskCnt;         // servo task general counter
int16_t ServoTaskDel;         // servo task delay skip timer
int16_t ServoTaskPnt;         // servo task pointer, default = 0
byte servoTgt;                // serevo target channel for future angle data.
bool ServoTgt10;              // == true if a 10mm cut-off point is detected
bool ServoTgtRec;             // == true if target angle is recorded during scanning
int16_t ServoUL;              // micro servo PWM upper limit
int16_t servoVal;             // PWM value applied to the micro servo
bool SpdChg;                  // == true if a wheel speed change is needed
int16_t SpeedInc;             // auto speed adjust increment, default is 4
int16_t SpdMax;               // maximum speed when moving
int16_t SpdMid;               // middle speed when moving
int16_t SpdMin;               // minimum speed when moving
int16_t SpdTop;               // current top speed when moving
int16_t Speed;                // default speed value
int16_t SpeedSrt;             // starting speed in AI mode
int16_t SpeedTgt;             // target speed figure, prevents sudden changes
int16_t Steer;                // turning factor, +ve right, -ve left
int16_t sw0Cnt;               // button switch counter
int16_t sw0DwnTime;           // button pressed down time
bool sw0LastState;            // previous state of button switch, HIGH/LOW
bool SW0_Nop;                 // == true to block SW0 read tasks, default = false
bool sw0State;                // state of read button switch pin
int16_t sw0Timer;             // timer used to detemine button sequences
int16_t sw1Cnt;               // button switch counter
int16_t sw1DwnTime;           // button pressed down time
bool sw1LastState;            // previous state of button switch, HIGH/LOW
bool SW1_Nop;                 // == true to block SW1 read tasks, default = false
bool sw1State;                // state of read button switch pin
int16_t sw1Timer;             // timer used to detemine button sequences
unsigned long t0,t1;          // millisecond code timers, report when t1 > t0
int16_t Task10ms;             // 10ms subtask timer
int16_t Task20ms;             // 20ms subtask timer
int16_t Task40ms;             // 40ms subtask timer
bool TgtOR;                   // == true if target was out of range
byte Tip;                     // > 0 if bot is tipping, 1 = pitch, 2 = yaw right, 4 = yaw left
byte TipLast;                 // previous Tip value
byte TipRnd;                  // random tip movement, 0 or 1
int16_t Turn_Cnt;             // counter timer used in turning
int16_t Turn_Time;            // counts time in 20ms when turning
byte Turn_Flg;                // set to IR_Mask when turning
int16_t Turning;              // direction of turn flag, 1 = Rht, -1 = Lft
int16_t Tx_PingCnt;           // 1 second Tx ping timer
bool USBPwr;                  // true if POST detects USB power voltage
bool WaitUp;                  // true if task is to wait for button release
int16_t WiFiCntC;             // C button counter for WiFi enable
int16_t WiFiConCnt;           // counter used in connection retry times
bool WiFiConnected;           // == true if a WiFi connection has been established
bool WiFiEn;                  // true when WiFi is enabled with prolonged 'Z' button activity
int16_t WiFiPing;             // Rx counter used to maintain connected state
int16_t WiFiRestCnt;          // counter used to goi to rest if no WiFi demands
bool WiFiRx;                  // true when a block of data has been received
long WiFiRxBytes;             // number of bytes received over the WiFi link
int16_t WiFiRxErr;            // number of bytes received over the WiFi link
int16_t WiFiTryCnt;           // WiFi try to connect count on error count
int16_t WiFiTryNum;           // WiFi try to connect total count
bool WiFiTx;                  // true when a block of data has been sent successfully
int16_t WiFiTx_CB;            // == 1 call back is clear to send, == -1 resend
int16_t WiFiTx_Chk;           // Tx checksum
int16_t WiFiTx_len;           // >0 when WiFi Tx buffer contains data to be sent
long WiFiTxBytes;             // number of bytes received over the WiFi link
int16_t WiFiTxErr;            // WiFi data error
int16_t WiFiTxErrCnt;         // WiFi Tx error count, leads to disconnection
int16_t WiiType;              // device type, 0 == Nunchuk, 1 == Classic
byte Z_Dn;                    // button pressed state; 0 == UP, >= 1 == Down

// ---------------------------------------------------------------------------------

void setup() {
  // Initialise I/O
  pinMode(LED_BUILTIN, OUTPUT); digitalWrite(LED_BUILTIN,HIGH);
  pinMode(FLftPwr, OUTPUT); digitalWrite(FLftPwr,LOW);  // sensor connected to ADC0
  pinMode(FCtrPwr, OUTPUT); digitalWrite(FCtrPwr,LOW);  // sensor connected to ADC3
  pinMode(FRhtPwr, OUTPUT); digitalWrite(FRhtPwr,LOW);  // sensor connected to ADC6
  pinMode(RRhtPwr, OUTPUT); digitalWrite(RRhtPwr,LOW);  // sensor connected to ADC7
  pinMode(RLftPwr, OUTPUT); digitalWrite(RLftPwr,LOW);  // sensor connected to ADC4
  pinMode(FLftADC,INPUT);  // ADC4
  pinMode(FCtrADC,INPUT);  // ADC7
  pinMode(FRhtADC,INPUT);  // ADC3
  pinMode(RRhtADC,INPUT);  // ADC6
  pinMode(RLftADC,INPUT);  // ADC0
  pinMode(33,INPUT);  // ADC5  battery divider
  pinMode(XSHUT,OUTPUT); digitalWrite(XSHUT,LOW);   // VL53L0X reset is active LOW
  pinMode(sw0Pin,INPUT_PULLUP);  // button switch sw0 Pin
  pinMode(sw1Pin,INPUT_PULLUP);  // button switch sw1 Pin

  // Allow allocation of all timers for servo library
  ESP32PWM::allocateTimer(0);
  ESP32PWM::allocateTimer(1);
  ESP32PWM::allocateTimer(2);
  ESP32PWM::allocateTimer(3);

  Wire.begin();

  Serial.begin(115200);   // baud rate for USB Rx/Tx comms
  Serial2.begin(115200);  // baud rate for LX-16A servo comms
  
  EEPROM.begin(EEPROM_SIZE);  // initialize EEPROM with predefined size
    
  // initialise WiFi link
  WiFi.mode(WIFI_STA);
  myMAC = WiFi.macAddress();
  WiFi.disconnect();
  
  // Set ESP-NOW Role as bidirectional
//  esp_now_set_self_role(ESP_NOW_ROLE_COMBO);

  // Init ESP-NOW and returns its status
  if (esp_now_init() != ESP_OK) {
    Serial.println("Error initializing ESP-NOW");
    return;
  }

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);
  
 
  // Register peer
  esp_now_peer_info_t peerInfo; //initialize and assign the peer information as a pointer to an addres
  memcpy(peerInfo.peer_addr, broadcastAddress, 6); //copy the value of  broadcastAddress with 6 bytes to peerInfo.peer_addr
  peerInfo.channel = 0;  //channel at which the esp talk. 0 means undefined and data will be sent on the current channel. 1-14 are valid channels which is the same with the local device 
  peerInfo.encrypt = false; //not encrypted

  //Add the device to the paired device list 
  if (esp_now_add_peer(&peerInfo) != ESP_OK){ 
    Serial.println("Failed to add peer");
    return;
  }

  
  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
  
  RunPOST();
  
  // set loop timers
  synchLoopTimers(0);
}

// ---------------------------------------------------------------------------------

void(* resetFunc) (void) = 0; //declare reset function at address 0

// ---------------------------------------------------------------------------------


void loop() {
  // main code which runs repeatedly as multi-tasking functions
  if (Reboot) {return;} // stop everything as a reset is pending

  if (LTOF_On > 0) LTOF_Read_Tasks();
  if (RangeNew) {if (LTOF_On == 5) {Servo_Task(); RangeNew = false;}} // micro servo movement tasks

  if (WiFiConnected) {
    if (WiFiRx) {readWiFiRx();} // buffer contains unread data so deal with it before adding new
  }

  readSerial();               // check serial port

  //###############################################################################
  //
  //  5 ms scanning tasks (200Hz)
  //
  //###############################################################################
  // These tasks are run at a high rate in order to detect ring edge collisions
  // the display ios locked in battle mode to prevent time lost to refresh
  // in non-battle mode the high rate will be disrupted by screen refresh, so we
  // add 5ms to the current time to avoid null readings on the sensors
  if (millis() >= next5ms) {next5ms = millis() + 5; readSensors();}


  //###############################################################################
  //
  //  10 ms non-critical subtasks (100Hz)
  //
  //###############################################################################
  // These tasks are run in a consequitive ripple to reduce the overhead on more
  // critical code. They are performed in reverse case order and will all complete
  // within the 10ms period
  if (Task10ms > 0) {
    switch(Task10ms) {
      case 1: if (LEDshow) {pixels.show(); LEDshow = false;} break;
      case 2: readSW1(); break;  // read right button switch SW1
    } Task10ms--; // count down the sub-tasks from 3 - 1 as we loop
  } else if (millis() >= next10ms) {
    next10ms += 10; Task10ms = 2;
    readSW0();  // read left button switch SW0 as priority task
  }


  //###############################################################################
  //
  //  20 ms Tasks (50Hz)
  //
  //###############################################################################
  // These tasks are run in a consequitive ripple to reduce the overhead on more
  // critical code. They are performed in reverse case order and will all complete
  // within the 20ms period
  if (Task20ms > 0) {
    switch(Task20ms) {
      case 1: if (MoveTask > 0) {MoveTasks();} break;
      case 2: if (MainTask > 0) {doMainTasks();} break;
      case 3: LED_Tasks(); break;     // various LED tasks
      case 4: calcAngles(); break;    // calculate yaw, pitch and turning
      case 5: AvgGyrAll(); break;     // average sensor Gyro values
      case 6: AvgAccelAll(); break;   // average sensor Accel values
      case 7: readAccelAll(); break;  // read MPU6050 accelerometers
    } Task20ms--; // count down the sub-tasks from 8 - 1 as we loop
  } else if (millis() >= next20ms) {
    next20ms += 20; Task20ms = 7;
    readGyrosAll(); // read MPU6050 gyroscopes as prioroty task
  }


  //###############################################################################
  //
  //  40 ms Tasks (25Hz)
  //
  //###############################################################################
  // These tasks are run in a consequitive ripple to reduce the overhead on more
  // critical code. They are performed in reverse case order and will all complete
  // within the 40ms period
  if (Task40ms > 0) {
    switch(Task40ms) {
      case  1: Display_40ms(); break;  // update display if necessary
      case  2: if (LTOF_On == 0) {Servo_Task();} break;    // micro servo movement tasks
      case  3: if (Ping > 0) {Ping--;} break;  // time-out data received flag
      case  4: // if not connected to WiFi run then try to connect every second
        if (!WiFiConnected) {
          // only try to connect WiFi if no 'Pings' are being received
          if (Ping == 0) {
            WiFiConCnt--; if (WiFiConCnt < 1) {
              WiFiConCnt = 50; WiFiTryToConnect();
//              Serial.println("Trying to connect WiFi");
            }
          }
        } else {
          // connceted to WiFi so read Wii data and respond
          readWii();
        } break;
      case  5: if (WiFiPing > 0) {
        WiFiPing--;
        if (WiFiPing < 1) {WiFiDisconnect();} // we have stopped receiving so disconnect WiFi
        } break;
    } Task40ms--;
  } else if (millis() >= next40ms) {
    next40ms += 40; Task40ms = 5;
  }


  //###############################################################################
  //
  //  40 ms Print Task - Variable
  //
  //###############################################################################
  // this code uses either the serial port or the WiFi link to send messages
  if (millis() >= print40ms) {
    print40ms = millis() + 40;
    if (!Echo) {PrintTx = "";}  // Echo is false so empty print buffer
    if (PrintTx.length() > 0) {
      // characters in string buffer so send some/all of them
      print40ms = millis() + 4; // shorten the delay for more data to send
      Tx_PingCnt = 0;           // reset the WiFi ping counter
      if (PrintTgt == 0) {
        // default uses the serial port
        if (Serial.availableForWrite() >= 64) {
          if (PrintTx.length() <= 64) {
            Serial.print(PrintTx);
            PrintTx = "";   // empty the buffer
          } else {
            Serial.print(PrintTx.substring(0,64));
            PrintTx = PrintTx.substring(64);
          }
        }
      } else {
        // text data has been received from WiFi so respond using ESP_NOW
        // ensure that we have had a call back from the previous frame
        if (WiFiConnected) {
          // only trannsmit PrintTx data if connected
          if (WiFiTx_CB == 1) {
            WiFiTx_CB = 0;  // clear the call back flag when sending
            if (PrintTx.length() <= 32) {
              WiFiTx_len = PrintTx.length();
              PrintTx.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = "";   // empty the buffer
            } else {
              WiFiTx_len = 32;
              Any$ = PrintTx.substring(0,WiFiTx_len);
              Any$.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = PrintTx.substring(32);
            }
          } else if (WiFiTx_CB == -1) {
            // an error has occured so resend the data
            if (WiFiTryCnt < 5) {
              WiFiTryCnt++;   // track the number of retries
              WiFiTx_CB = 0;  // clear the call back flag when sending
//              t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
            } else {
              // too many resends so disconnect WiFi link
              WiFiDisconnect();
              Serial.println("Failed to Tx!");
            }
          }
        } else {
          // not connected so clear PrintTxdata
          PrintTx = "";
        }
      }
    }
  }


  //###############################################################################
  //
  //  200 ms Tasks (5Hz)
  //
  //###############################################################################
  if (millis() >= next200ms) {
    next200ms += 200;
    // test for RESET condition
    if (sw1State == LOW) {
      ResetCnt++; if (ResetCnt >= 15) {
        // button held down for >= 3 seconds so force a soft reset
//        Serial.println("Reset");
        Display_Reboot(); delay(1000);
        resetFunc();    //call reset
      }
//      Serial.println("ResetCnt " + String(ResetCnt));
    } else {ResetCnt = 0;}  // clear the timer
  }

  //###############################################################################
  //
  //  1 sec Tasks
  //
  //###############################################################################
  if (millis() >= nextSec) {
    // toggel built-in LEDS every second to show that the code is still working
    nextSec += 1000;
    LED_ON = !LED_ON;
    if (LED_ON) {
      digitalWrite(LED_BUILTIN,HIGH);
    } else {
      digitalWrite(LED_BUILTIN,LOW);
    }
  }
}

// ---------------------------------------------------------------------------------

void loopDelay(long zDelay) {
  // called to run the main loop for a zDelay period in ms before returning
  // note the calling function must block being called from loop() to prevent stack
  // overflow conditions
  zDelay += millis(); // set the exit time
  while (millis() < zDelay) {
    loop();   // run tasks in the main loop
    yield();  // pass control to background functions
  }
}

// --------------------------------------------------------------------------------

void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) {
  // Callback when data is received
  // we copy the data into a buffer and set a flag so that the main loop will act
  // to handle it
  while (WiFiRx) {
    // buffer contains unread data so deal with it before adding new
    readWiFiRx();
  }

  memcpy(&Rx_Buff, incomingData, len);

  // incoming data will always be two or more characters, ie. text == $+++
  if (len < 2) {Serial.println("Rx Error - no Rx data"); return;}
  
  if (Rx_len > 0) {Serial.println("Rx Error - buffer in use");} // this should not happen

  // if connected Rx data will have a checksum so check it first
  Rx_len = len - 1; Rx_Pnt = 0;
  RxGetChecksum(Rx_len);  // generate a checksum from Rx data
  if (Rx_Chk != Rx_Buff.ESPdata[len - 1]) {
    Serial.println("Rx Error - checksum");
  }
  // set flags before exiting
  WiFiRxBytes += Rx_len; Rx_Task = 0; WiFiRx = true;
  WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
//    Serial.print("Bytes received: "); Serial.println(len);
}

// --------------------------------------------------------------------------------

void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  // Callback when data is sent, determines success or failure
//  t1 = micros();
//  Serial.print("Send Status: ");
  if (WiFiConnected) {
    // the link is already active so handle normal success/failure scenarios
    if (status == 0){
//      Serial.println("Success " + String(t1-t0));
      WiFiTx = true;    // flag WiFi Tx success
      WiFiTx_CB = 1;    // set flag to indicate that a call back state has occured
      WiFiTxErrCnt = 0; // clear the error flag in case of a glitch
      WiFiPing = 50;    // set link time-out to 2 seconds
    } else {
//      Serial.println("Fail " + String(t1 - t0));
      WiFiTx = false;   // flag WiFi Tx failure
      WiFiTx_CB = -1;   // set flag to indicate that a resend is required
      WiFiTxErrCnt++;
    }
  } else {
    // not connected so we must have tried to connect and may get an error
    if (status == 0){
      // successful so enter the connected state
//      Serial.println("Connected!");
      WiFiConnected = true;
      WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
      WiFiTxErrCnt = 0; // reset error count for this connection
    } else {
//      Serial.println("Status = " + String(status));
      WiFiTxErrCnt++;
    }
    WiFiTx_CB = 1;
  }
}

// ---------------------------------------------------------------------------------

void RunPOST() {
  // start-up initialisation processes  
  setDefaults();  // reset variables
  Serial.println("\n\nSumo Robot v0.0\n");
  setMainTask(0); // ensure wheels are stationary
  readBattery(0); // take a battery reading

  // Initialise display
  Display_Init();
  Display_Intro();
  Serial.println("Display initialised");

  // initilaise RGB LED array
  pixels.begin();
  pixels.fill(0,0,5);  // strip.fill(color, first, count)
  pixels.show(); // This sends the updated pixel color to the hardware.
  NeoPixel_Clear_Array(); NeoPixel_ShowArray(true);

  // get Flash data
  if (EEPROM_Checksum()) {
    // valid EEPROM checksum
    Serial.println("EEPROM data loaded");
    EEPROM_Read();
  } else {
    // invalid EEPROM checksum
    Serial.println("EEPROM Checksum Error!");
    Serial.println("Writing default values...");
    EEPROM_Write();
  }
  reportCalData();
  GyXOff = 0; GyYOff = 0; GyZOff = 0;     // clear gyro offsets, they are auto calibrated

  // centre the servo
  microServo.setPeriodHertz(50);          // Standard 50hz servo
  ServoON(); delay(500); ServoOFF();
  Serial.println("Centred servo");

  // initialise the accelerometer
  initialiseMPU();

  // initialise the LTOF laser ranger
  VL53L0X_ON();   // turn on LTOF ranger
  VL53L0X_OFF();  // turn on LTOF ranger
  Serial.println("LTOF initialised");

  // set up battery quick averaging
  readBattery(0); BatAvg = BatVol; BatSum = BatVol * 20;
  AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
  if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
  if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
  Serial.println("Battery " + String(float(BatAvg)/BatCal) + "v\t" + String(AnyLong) + "%");
  randomSeed(BatVol); // use the battery voltage as the random seed factor

  Serial.println("POST completed\n\n");
  delay(1000); // show SUMO logo a bit longer
}

// ---------------------------------------------------------------------------------

void setDefaults() {
  // load default values
  AcAvgDiv = 0;           // dividing factor used in averaging
  AcAvgDivMax = 250;      // max dividing factor for averaging
  AcXAvg = 0;             // averaged AcX value to reduce noise
  AcXOff = 0;             // offset to be added to raw X-acceleration value
  AcYAvg = 0;             // averaged AcX value to reduce noise
  AcYOff = 0;             // offset to be added to raw Y-acceleration value
  AcZAvg = 0;             // averaged AcX value to reduce noise
  AcZOff = 0;             // offset to be added to raw Z-acceleration value
  ADC = 0;                // IR sensor pointer, 0 - 4
  ADC_Cnt = 40;           // timer counter used in IR sensor reading
  ADC_Fast = false;       // IR sensor read flag; fast == true
  ADC_Rec = 0;            // ADC sensor channel to be recorded 0 - 4
  ADC_RecPnt = 0;         // ADC data array pointer used in storing data
  AI_ATTACK_EN = false;   // if == true it enables AI attack mode
  AI_EDGE_ON = false;     // if == true AI performs edge detection
  AI_IR_ON = false;       // if == true AI turns ON IR sensor scanning
  AI_LED_Task = -20;      // default AI LED task when in IR mode
  AI_LTOF_ON = false;     // == true enables LTOF servo scanner in AI mode
  AI_REPORT = true;       // if == true then report messages
  AI_Sub = 0;             // initial AI sub-task
  AI_TIP_ON = false;      // == true enables AI tip responses
  AI_Title$ = "";         // title used in AI display logo
  AI_TurnCnt = 0;         // timer counter used in AI steering
  Attack = false;         // AI task attacking mode
  BatAvg = 4095;          // preload averaging function with high value
  BatAvgMax = 0;          // maximum average value in this period
  BatCnt = 0;             // pointer to array values and critical timer
  BatCritCnt = 0;         // timer used in baterycritical testing
  BatDel = 50;            // set a 10 second delay in 'life' prediction
  BatDP = 0;              // battery data pointer
  BatLast = 0;            // previous battery average reading
  BattNow = false;        // == true causes battery reading
  BatPwr = false;         // true if battery supply detected, false if USB assumed
  BatRead = 8;            // timer counter used in reading battery
  BatRep = 0;             // time when to report battery reading over serial link
  BatSum = 72320;         // preload averaging sum with 20x max value
  BatT0 = 0;              // time of first battery measurement in life prediction forecast
  BatTime$ = "--:--";     // time left displayed
  BatV0 = 0;              // reference voltage used in prediction forecast
  BatV1 = 0;              // lowest voltage used in prediction forecast
  BatVol = 0;             // instantaneous battery voltage
  C_Cnt = 0;              // counter used in 'C' button detection
  C_Dn = 0;               // counter for 'C' button down period
  CountDown = 0;          // seconds count down used by AI
  cmdMode = ' ';          // command mode
  cmdSgn = 1;             // cmdVal sign, 1 or -1
  cmdType = ' ';          // command mode type
  cmdVal = 0;             // value associated with a cmdType
  CZ = 0;                 // received button values
  CZ_Wait = false;        // flag used for button release wait function
  DispBack = false;       // flag used to draw a back arrow on the display
  DispChng = false;       // flag invokes initialisation code
  DispClr = false;        // set == true to clear the memory contents
  DispCnt = 0;            // counter used in display updating
  DispDel = 0;            // >0 display delay counter, stops updates over I2C
  DispDisp = false;       // set = true to immediately display memory contents
  DispMenu = 0;           // group type, 0 = sensors, 1 = tasks, etc
  DispMode = 0;           // the type of display for a given mode
  DispNOP = false;        // set == true to stop display I2C updates
  DispNow = 0;            // if > 0 display special modes
  DispPnt = 0;            // points at the current display mode, default = 0 batteries
  Echo = true;            // if true send print messages
  ESC0 = false;           // SW0 ESC exit current Main Task flag
  ESC1 = false;           // SW1 ESC exit current Main Task flag
  GyAvgDiv = 0;           // dividing factor used in averaging
  GyCal = 250;            // counter used to determine gyro calibration period
  GyX = 0;                // MPU raw X-gyroscope value
  GyXAvg = 0;             // averaged GyX value to reduce noise
  GyXOff = 0;             // offset to be added to raw X-gyroscope value
  GyXOffCnt = 0;          // offset counter for gyro centering
  GyXSum = 0;             // averaging accumulator
  GyY = 0;                // MPU raw Y-gyroscope value
  GyYAvg = 0;             // averaged GyY value to reduce noise
  GyYOff = 0;             // offset to be added to raw Y-gyroscope value
  GyYOffCnt = 0;          // offset counter for gyro centering
  GyYSum = 0;             // averaging accumulator
  GyZ = 0;                // MPU raw Z-gyroscope value
  GyZAvg = 0;             // averaged GyZ value to reduce noise
  GyZOff = 0;             // offset to be added to raw Z-gyroscope value
  GyZOffCnt = 0;          // offset counter for gyro centering
  GyZSum = 0;             // averaging accumulator
  I2CError = true;        // used to detect I2C transmission errors
  IR_BaseF = true;        // base colour, true == white, false == black
  IR_Mask = 0;            // a bit mask which identifies edge detection
  JoyDL = 0;              // direction 1 counter
  JoyX = 128;             // received value, default
  JoyY = 128;             // received value, default
  L_Dst = 0;              // distance travelled in 0.01mm increments
  LED_Del = 0;            // LED Task delay, adds pauses into the processes
  LED_ON = false;         // state of LED pin HIGH/LOW
  LED_Nop = false;        // == true to block eye LED tasks, default = false
  LEDshow = false;        // LED show flag for mouth LEDs
  LED_SubTask = 0;        // LED sub-task pointer
  LED_SwDwn = false;      // == true if a button switch has been pressed
  LED_Task = -1;          // LED task pointer, -ve values reset variables
  LEDVal = 0;             // values used in LED tasks
  L_Spd = 0;              // speed value sent to left LX-16A servo
  LTOF_On = 0;            // > 0 when laser ranging is enabled, default is 0 == OFF
  LX16Phase = 0;          // either 0 or 1 for LX16 up/down speed
  LX16Spd = 0;            // speed value sent tot he servo, +/- 0 - 1000
  LX16Tgt = 1;            // servo target number for data transfer; default = 1
  MainCnt = 0;            // Main Task general counter/timer
  MainDel = 0;            // Main task delay counter, if > 0 holds off tasks
  MainNext = 0;           // next MainTask to be performed after a delay
  MainSub = 0;            // Main Task sub-task
  MainTask = 0;           // main task pointer
  MPUCal = false;         // == true during calibration
  myMAC = "";             // MAC address if this device
  MoveState = 0;          // move state machine pointer
  overRun = 0;            // counter used to prevent undemanded movement
  Ping = 0;               // 'ping' counter
  PitchAcc = 0.0;         // pitch angle in degrees base on accelerometer
  PitchGyr = 0.0;         // pitch angle in degrees base on gyro + acc
  PrintTx = "";           // clear printed strings
  PrintTgt = 0;           // 0 == serial port, 1 == WiFi ESP-NOW
  RangeLast = 600;        // 16-bit range value
  R_Dst = 0;              // distance travelled in 0.01mm increments
  readWiiCall = false;    // flag prevents call stack overflow
  Reboot = false;         // set true when reboot is in progress to stop all tasks
  ResetCnt = 0;           // timer used to invoke soft RESET
  RndDel = 0;             // random driving task skip counter/timer
  RndTask -1;             // tasks performed by random driving
  R_Spd = 0;              // speed value sent to right LX-16A servo
  Rx_Chk = 0;             // Rx checksum
  RxCZ = 0;               // buffered received CZ value
  RxJoyX = 0;             // buffered received JoyX value
  RxJoyY = 0;             // buffered received JoyY value
  Rx_len = 0;             // length of WiFi received data block
  RxRec = false ;         // set true when a valid frame of data is received
  RxState = 0;            // receiver state machine state
  Rx_Task = 0;            // pointer user by Rx state machine
  RxVal = -1;             // value received from serial Rx
  ServoAng = 90;          // micro-servo angle; default = 90
  ServoAngMax = ServoAngMaxDef; // micro-servo max angle; default = 130
  ServoAngMin = ServoAngMinDef; // micro-servo max angle; default = 50
  ServoCtr = ServoCtrDef; // centre Cal. PWM value for microservo
  ServoEn = false;        // == true when servo is attached
  ServoLL = ServoMin;     // micro servo PWM lower limit
  ServoMax = ServoMaxDef; // maximum Cal. PWM value for microservo
  ServoMin = ServoMinDef; // minimum Cal. PWM value for microservo
  servoPhase = 0;         // either 0 or 1 for up/down angle change
  ServoTaskDel = 0;       // servo task delay skip timer
  ServoTaskPnt = 0;       // servo task pointer, default = 0
  servoTgt = 0;           // serevo target channel for future angle data.
  ServoUL = ServoMax;     // micro servo PWM upper limit
  servoVal = ServoCtr;    // PWM value applied to the micro servo
  SpdTop = 0;             // current top speed when moving
  SpeedInc = 4;           // auto speed adjust increment, default is 4
  SpeedSrt = 0;           // starting speed in AI mode
  Steer = 0;              // turning factor, +ve right, -ve left
  sw0Cnt = 0;             // button switch counter
  sw0DwnTime = 0;         // button pressed down time
  sw0LastState = HIGH;    // previous state of button switch, HIGH/LOW
  SW0_Nop = false;        // == true to block SW0 read tasks, default = false
  sw0State = HIGH;        // state of read button switch pin
  sw0Timer = 0;           // timer used to detemine button sequences
  sw1Cnt = 0;             // button switch counter
  sw1LastState = HIGH;    // previous state of button switch, HIGH/LOW
  SW1_Nop = false;        // == true to block SW1 read tasks, default = false
  sw1State = HIGH;        // state of read button switch pin
  sw1Timer = 0;           // timer used to detemine button sequences
  t0 = 0;                 // microsecond timer used for code speed checks
  t1 = 0;                 // microsecond timer used for code speed checks
  Task10ms = 0;           // 10ms subtask timer
  Task20ms = 0;           // 20ms subtask timer
  Task40ms = 0;           // 40ms subtask timer
  TgtOR = true;           // == true if target was out of range
  Tip = 0;                // > 0 if bot is tipping, 1 = pitch, +/-2 = yaw
  Turn_Flg = 0;           // set to IR_Mask when turning
  Turn_Cnt = 0;           // counter timer used in turning
  Turning = 0;            // direction of turn flag, 1 = Rht, -1 = Lft
  Tx_PingCnt = 0;         // 1 second Tx ping timer
  UpRht = true;           // set == true by AcZ when the robot is the correct way up
  USBPwr = false;         // true if POST detects USB power voltage
  WaitUp = false;         // true if task is to wait for button release
  WiFiCntC = 0;           // C button counter for WiFi enable
  WiFiConCnt = 0;         // counter used in connection retry times
  WiFiConnected = false;  // == true if a WiFi connection has been established
  WiFiEn = false;         // true when WiFi is enabled with prolonged 'Z' button activity
  WiFiPing = 0;           // Rx counter used to maintain connected state
  WiFiRx = false;         // true when a block of data has been received
  WiFiRxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiRxErr;              // number of receive errors in the WiFi link
  WiFiTryCnt = 0;         // WiFi try to connect count
  WiFiTryNum = 0;         // WiFi try to connect total count
  WiFiTx = false;         // true when a block of data has been sent successfully
  WiFiTxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiTx_CB = 1;          // == 1 call back is clear to send, == -1 resend
  WiFiTx_Chk = 0;         // Tx checksum
  WiFiTx_len = 0;         // >0 when WiFi Tx buffer contains data to be sent
  WiFiTxErr = 0;          // WiFi data error
  WiFiTxErrCnt = 0;       // WiFi Tx error count, leads to disconnection
  WiiType = 0;            // device type, 0 == Nunchuk, 1 == Classic
  YawAcc = 0.0;           // yaw angle in degrees based on accelerometers
  YawGyr = 0.0;           // yaw angle in degrees based on gyro + acc
  Z_Dn = 0;               // button pressed state; 0 = UP, 1 = Down
}

// ---------------------------------------------------------------------------------

void synchLoopTimers(unsigned long zAdvance) {
  // reset main loop timers
  // a 2ms phase offset is introduced to prevent simultaneous triggering
  DistanceMs = millis();  // reset the distance timer
  if (zAdvance == 0) {
    // setting up fresh loop timers
    next5ms = millis() + 11; // reset 5ms loop timer with 1ms phase to 10ms timer
    next10ms = millis() + 10;
    next20ms = millis() + 22; // reset 20ms loop timer with 2ms phase to 10ms timer
    next40ms = millis() + 43; // reset 40ms loop timer with 3ms phase to 10ms timer
    print40ms = millis() + 44; // reset 40ms printer loop timer with 4ms phase to 10ms timer
    next200ms = millis() + 205; // reset 200ms loop timer with 5ms phase to 10ms timer
    nextSec = millis() + 1006; // reset 1sec loop timer with 6ms phase to 10ms timer
  } else {
    // advancing loop timers
    next5ms += zAdvance;
    next10ms += zAdvance;
    next20ms += zAdvance;
    next40ms += zAdvance;
    print40ms += zAdvance;
    next200ms += zAdvance;
    nextSec += zAdvance;
  }
}

// --------------------------------------------------------------------------------
