
// --------------------------------------------------------------------------------

void AvgAccelAll() {
  // average the raw accelerometer values over a given number of values
  AcXSum += AcX; AcYSum += AcY; AcZSum += AcZ;
  if (AcAvgDiv < AcAvgDivMax) {
    // increase the scope of the averaging factor if necessary
    // note at 20ms sample, AvgDiv = 50 == 1 second rolling average
    AcAvgDiv++;
  } else {
    // once the accumulator has the correct number of samples roll out the average
    AcXSum -= AcXAvg; AcYSum -= AcYAvg; AcZSum -= AcZAvg;
  }
  AcXAvg = AcXSum / AcAvgDiv;
  AcYAvg = AcYSum / AcAvgDiv;
  AcZAvg = AcZSum / AcAvgDiv;
}

// ---------------------------------------------------------------------------------

void AvgGyrAll() {
  // called as a 20ms subtask, keeps a rolling average of the gyroscopes
  GyXSum += GyX; GyYSum += GyY; GyZSum += GyZ;
  if (GyAvgDiv < GyAvgDivMax) {
    // increase the scope of the averaging factor if necessary
    // note at 20ms sample, AvgDiv = 50 == 1 second rolling average
    GyAvgDiv++;
  } else {
    // once the accumulator has the correct number of samples roll out the average
    GyXSum -= GyXAvg; GyYSum -= GyYAvg; GyZSum -= GyZAvg;
  }

  GyXAvg = GyXSum/GyAvgDiv;
  GyYAvg = GyYSum/GyAvgDiv;
  GyZAvg = GyZSum/GyAvgDiv;

  // initially the Gyro is not calibrated, so we run a timer before using its output
  // as an offset value
  if (GyCal > 0) {
    GyCal--; 
    if (GyCal < 1) {
      // calibration period complete so use offsets determined by averaging
      GyXOff = -GyXAvg;
      GyYOff = -GyYAvg;
      GyZOff = -GyZAvg;
      PitchGyr = PitchAcc;  // set pitch gyro using Acc angle
      RotGyr = 0.0;         // zero the rotation gyro
      YawGyr = YawAcc;      // set yaw gyro using Acc angle
    }
  }
}

// ---------------------------------------------------------------------------------

void calcAngles() {
  // called from main loop every 20ms to calculate the yaw, pitch and turn angles
  // PitchAcc +ve tip forward, -ve tip backwards
  // YawAcc +ve tip right, -ve tip left
  // we want to reverse the sign of these values
  Tip = 0;  // reset the tipping flag
  PitchAcc = -57.2958 * asin((float)AcZ/8192.0);  // angle in radians
  YawAcc = -57.2958 * asin((float)AcX/8192.0);    // angle in radians
  if (MPUCal) {PitchAcc = 0.0; YawAcc = 0.0;}     // in calibration mode

  // only do these calculations once the calibration process is completed
  if (GyCal > 0) {
    // during calibration gyros are locked to zero
    PitchGyr = 0.0; YawGyr = 0.0; RotGyr = 0.0;
  } else {
    // gyros are set at +/-250 deg/sec for 16-bit ADC = 32,767 FSD
    // to convert gyro value to deg/sec we divide by 131.068 (ie. 32,767/250)
    // we have recorded time period in GyTD in milliseconds, between readings
    // so angle is = GyAng * (GyTD/1000)/131.068
    // GyX = pitch, +ve tipping back, -ve tipping forward
    // GyY = rotate round
    // GyZ = yaw, +ve tipping left, -ve tipping right 
    PitchGyr -= (float)GyX * (float)GyTD / 131068.0;
    // use 5% accelerometer angle difference to correct any drift
    PitchGyr = PitchGyr + ((PitchAcc - PitchGyr) * 0.05);
    // limit the angle
    if (PitchGyr > 360.0) {PitchGyr -= 360.0;}
    else if (PitchGyr < -360.0) {PitchGyr += 360.0;}
    PitchGyrInt = int(PitchGyr);  // grab integer form for faster decision making
  
    YawGyr += (float)GyZ * (float)GyTD / 131068.0;
    // use 5% accelerometer angle difference to correct any drift
    YawGyr = YawGyr + ((YawAcc - YawGyr) * 0.05);
    // limit the angle
    if (YawGyr > 360.0) {YawGyr -= 360.0;}
    else if (YawGyr < -360.0) {YawGyr += 360.0;}
    YawGyrInt = int(YawGyr);  // grab integer form for faster decision making
  
    // determine tipping condition
    // we can respond to tipping backwards by going backwards and turning
    // tipping sideways has a great angle as the Yaw gyro is affected by the 3 deg lean
    // 1 = pitch tipping backwards
    // 2 = yaw tipping right
    // 3 = yaw tipping left
    if (PitchGyrInt >= 5) {Tip += 1;}       // tipping backwards
    if (YawGyrInt >= 15) {Tip += 2;}        // tipping to right
    else if (YawGyrInt <= -15) {Tip += 2;}  // tipping to left
  
    // perform long average offset adjustment to improve accuracy
    if (GyCal == 0) {
      if (PitchGyr < PitchAcc) {
        GyXOffCnt--;
        if (GyXOffCnt < -100) {GyXOffCnt = 0; GyXOff--;}
      } else {
        GyXOffCnt++;
        if (GyXOffCnt > 100) {GyXOffCnt = 0; GyXOff++;}
      }
      if (YawGyr < YawAcc) {
        GyZOffCnt--;
        if (GyZOffCnt < -100) {GyZOffCnt = 0; GyZOff++;}
      } else {
        GyZOffCnt++;
        if (GyZOffCnt > 100) {GyZOffCnt = 0; GyZOff--;}
      }
    }
  
    RotDif = (float)GyY * (float)GyTD / 131068.0;
  //  Serial.println(String(fabs(RotDif)) + "\t" + String(RotCnt));
    RotGyr += RotDif;
    // if some rotation is detected then set a timer, if not then
    // ignore the drift by setting the angle to zero
    if (fabs(RotDif) > 0.01) {RotCnt = 150;} else if (RotCnt == 0) {RotGyr = 0.0;}
    // switch the angle to zero after 3 sec time-out
    if (RotCnt > 0) {
      // we are or have been rotating
      RotCnt--; if (RotCnt < 1) {RotGyr = 0.0;} // zero the rotation gyro
    } else {
      // we appear to have stopped or rotation is small
      // in this condition perform long averaging to improve accuracy
      if (GyY > 0) {
        GyYOffCnt++;
        if (GyYOffCnt > 100) {GyYOffCnt = 0; GyYOff--;}
      } else {
        GyYOffCnt--;
        if (GyYOffCnt < -100) {GyYOffCnt = 0; GyYOff++;}
      }
    }
    // limit the angle
    if (RotGyr > 360.0) {RotGyr -= 360.0;}
    else if (RotGyr < -360.0) {RotGyr += 360.0;}
    RotGyrInt = int(RotGyr);
  }


//  Serial.print(AcZ); Serial.print("\t");
//  Serial.print(AcX); Serial.print("\t");
//  Serial.print(GyZ); Serial.print("\t");
//  Serial.println(GyX);

//  Serial.print(PitchAcc); Serial.print("\t");
//  Serial.print(YawAcc); Serial.print("\t");
//  Serial.print(PitchGyr); Serial.print("\t");
//  Serial.println(YawGyr);
}

// ---------------------------------------------------------------------------------

void Cal_SetServoOE(int zVal) {
  // manually set the servo attachment
  if (zVal != 1) {
    // enabling servo
    ServoON(); PrintTx += "Servo [ON]\n";
  } else {
    // disabling micro servo
    ServoOFF(); PrintTx += "Servo [OFF]\n";
  }
}

// ---------------------------------------------------------------------------------

void Cal_SetServoValue() {
  // set the micro servo value LL - UL and report back
  if ((cmdVal >= ServoLL) && (cmdVal <= ServoUL)) {
    // in range value received
    servoVal = cmdVal;
    if (!ServoEn) {ServoON();}
    microServo.writeMicroseconds(servoVal);
    PrintTx += "SV=" + String(servoVal) + "\n";
  } else {
    PrintTx += "SV ERROR!\n";
  }
}

// ---------------------------------------------------------------------------------

void Cal_SetServoVLL() {
  // store the lower limt servo PWM value
  ServoLL = cmdVal;
}

// ---------------------------------------------------------------------------------

void Cal_SetServoVUL() {
  // store the upper limt servo PWM value
  ServoUL = cmdVal;
}

// ---------------------------------------------------------------------------------

void clearADCdata() {
  // called to clear the contents of the recorded sensor data
  for (int zP = 0; zP < 40; zP++) {
    ADC_Data[zP] = -1;  // set data to -ve out of limit value, so it is not plotted
  }
}

// ---------------------------------------------------------------------------------

void clearRangeData() {
  // called to reset stored range data
  int zPMax = 1 + (ServoAngMaxDef - ServoAngMinDef);
  for (int zP = 0; zP < zPMax; zP++) {
    RangeData[zP] = RangeMax;  // set data to max range value
  }
}

// ---------------------------------------------------------------------------------

void decodeKey(int zkeyVal) {
  // decodes zkeyVal and excutes commands on '.'
  // every time we receive a character set response to 10ms later, just in case
  // there is another packet behind it coming from the controller
  // when Ping data is received set AutoOffTimer and eye LEDs to red
  next40ms = millis() + 10;
  if (zkeyVal == 10) {return;}
  if (zkeyVal == 13) {return;}
  keyChar = char(zkeyVal);
  switch (keyChar) {
    case '.': doCmd(); return;          // command terminator
//    case '+': cmdVal = -1; return;      // set flag for ML+. command
    case '~': // connected to OLED Mirror app via USB or WiFi
      if (!DispMon) {DispMon = true; AI_REPORT = false; Display_Mirrored();} // display mirrored message
      DispTx = 50; return;
    case '-': cmdSgn = -1; return;      // numeric sign
//    case '!': SetMainTask(98); return;  // return to 'Rest' position
//    case '#': Ping = 50; AutoOffTimer = 0;  LEDTaskPnt = 3; return;   // 'ping' sent from application, disables auto-rest function
    case '*': cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1; return; // ESC abort command
    case '0': extendCmdVal(0); return;
    case '1': extendCmdVal(1); return;
    case '2': extendCmdVal(2); return;
    case '3': extendCmdVal(3); return;
    case '4': extendCmdVal(4); return;
    case '5': extendCmdVal(5); return;
    case '6': extendCmdVal(6); return;
    case '7': extendCmdVal(7); return;
    case '8': extendCmdVal(8); return;
    case '9': extendCmdVal(9); return;
  }
  if (cmdMode == ' ') {
    // test for new Command Mode char?
    // only accept certain characters as valid commands, or ignore them
    // delay battery update by 10 seconds whilst receiving commands
    BatRep = millis() + 10000;
    switch (keyChar) {
      // check for valid mode and convert lower-case
      case 'd': cmdMode = 'D'; break;
      case 'D': cmdMode = 'D'; break;
      case 'l': cmdMode = 'L'; break;
      case 'L': cmdMode = 'L'; break;
      case 's': cmdMode = 'S'; break;
      case 'S': cmdMode = 'S'; break;
      case 't': cmdMode = 'T'; break;
      case 'T': cmdMode = 'T'; break;
    } cmdType = ' '; cmdVal = 0;
  } else {
    // test for Command Type char?
    cmdType = keyChar;
    switch (keyChar) {
      // check for lower-case and convert to upper-case
      case 'l': cmdType = 'L'; break;
      case 'm': cmdType = 'M'; break;
      case 'o': cmdType = 'O'; break;
      case 'r': cmdType = 'R'; break;
      case 't': cmdType = 'T'; break;
      case 'u': cmdType = 'U'; break;
      case 'v': cmdType = 'V'; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void disableWiFi() {
  // when either a switch is pressed or WiFi disable is required
  WiFiEn = false; WiFiCntC = 0; WiFiRestCnt = 0; C_Dn = 8; Z_Dn = 8;
}

// ---------------------------------------------------------------------------------

void doCmd() {
  // a '.' has been received on the serial port so execute command if valid
  // Commands:
  // DMxxyy.    - display monitor app mouse down at xx%,yy%, 0 - 9999
  // DM-.       - display monitor app mouse up.
  // LTnn.      - set the LX-16A target channel for future data. Default = 1.
  // LVnn.      - set the target LX-16A servo motor value
  // R.         - report XDebug data over the serial link
  // RL.        - reports XDebug lables for RPnn over the serial link
  // RPnn.      - report nn XDebug data over the serial link
  // SLnn.      - set the target servos lower limit
  // SOn.       - set attached state, 0 = enable, 1 = disable
  // SRn.       - set the state of the echo flag, 0 = OFF, 1 = ON
  // STnn.      - set the servo target channel for future angle data. Default = 0.
  // SUnn.      - set the target servos upper limit
  // SVnn.      - set the target servo PWM value
  // TMnn.      - invoke special test modes

  cmdVal*= cmdSgn;  // correct cmdVal with the sign value
  
  switch (cmdMode) {
    case ' ': break;
    case 'D':
      switch (cmdType) {
        case 'M':
          // react to display monitor mouse click
          if (cmdSgn < 0) {DMsw0 = true; DMsw1 = true;}
          else {cmdVal /= 100; if (cmdVal < 50) {DMsw0 = false;} else {DMsw1 = false;}}
          break;
      } break;
    case 'L':
      switch (cmdType) {
        case 'T':
          // set the LX-16A target channel for future data. Default = 1.
          PrintTx += "LT" + String(cmdVal) + ".\n";
          LX16Tgt = cmdVal; break;
        case 'V': SetLX16Value(); break;
      } break;
//    case 'R':
//      switch (cmdType) {
//        case ' ': 
//          // report the same set of values twice as WiFi data can easily be
//          // corrupted. Therefore two consequitives messages must be received
//          // for the PID Controller graphing function to work
//          // the GR count is used to inhibit other forms of reporting, like battery
//          RPT = 0; DebugReport(RPnn); DebugReport(RPnn); GR = 25; break;
//        case 'L':
//          // report trace labels for PID Graph app
//          RPT = 1; DebugReport(RPnn); break;
//        case 'P':
//          RPnn = cmdVal; 
//          PrintTx += "RP" + String(RPnn) + "\n"; break;
//      } break;
    case 'S':
      switch (cmdType) {
        case 'L': Cal_SetServoVLL(); break;
        case 'O': Cal_SetServoOE(cmdVal); break;
        case 'R':
          // set the state of the Echo flag
          if (cmdVal > 0) {Echo = true; PrintTx += "Echo ON\n";} else {Echo = false;}
          break;
        case 'T':
          // set the servo target channel for future angle data. Default = 0
          PrintTx += "ST" + String(cmdVal) + ".\n";
          servoTgt = cmdVal; break;
        case 'U': Cal_SetServoVUL(); break;
        case 'V': Cal_SetServoValue(); break;
      } break;
    case 'T':
      switch (cmdType) {
        case 'M': TestModeRun(cmdVal); break; // invoke special test mode
      }
      break;
    default:
      PrintTx += "Unknown cmd: " + cmdMode + cmdType + String(cmdVal) + "\n";
      break;
  }
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdSgn = 1;
}

// ---------------------------------------------------------------------------------

bool EEPROM_Checksum() {
  // reads the contents of the EEPROM data and confirms valid checksum
  // checksum is stored 1st as 2-byte value (LSB/MSB)
  // data is stored as 2-byte integer variables (LSB/MSB)
  int16_t zCheck = EEPROM.read(0) + (EEPROM.read(1)<<8);  // read the checksum
//  Serial.print("LSB:" + String(EEPROM.read(0)) + "\t");
//  Serial.println("MSB:" + String(EEPROM.read(1)));
  int16_t zSum = 0; // clear the sum
  for (int zF = 2;zF < EEPROM_SIZE;zF++) {
    zSum += EEPROM.read(zF);  // accumulate each byte of data
//    Serial.println(zSum);
  }
  // Serial.println("Checksum = " + String(zCheck));
  // Serial.println("Data Sum = " + String(zSum));
  
  if (zCheck == zSum) {return true;} else {return false;}
}

// ---------------------------------------------------------------------------------

void EEPROM_Read() {
  // read calibration data from EEPROM flash memory
  // normally called after a valid checksum test has been performed
  // read IR black thresholds
  AvBlk[0] = EEPROM.read( 2) + (EEPROM.read( 3)<<8);
  AvBlk[1] = EEPROM.read( 4) + (EEPROM.read( 5)<<8);
  AvBlk[2] = EEPROM.read( 6) + (EEPROM.read( 7)<<8);
  AvBlk[3] = EEPROM.read( 8) + (EEPROM.read( 9)<<8);
  AvBlk[4] = EEPROM.read(10) + (EEPROM.read(11)<<8);

  // read IR white thresholds
  AvWht[0] = EEPROM.read(12) + (EEPROM.read(13)<<8);
  AvWht[1] = EEPROM.read(14) + (EEPROM.read(15)<<8);
  AvWht[2] = EEPROM.read(16) + (EEPROM.read(17)<<8);
  AvWht[3] = EEPROM.read(18) + (EEPROM.read(19)<<8);
  AvWht[4] = EEPROM.read(20) + (EEPROM.read(21)<<8);

  // read accelerometer offsets
  AcXOff = EEPROM.read(22) + (EEPROM.read(23)<<8);
  AcYOff = EEPROM.read(24) + (EEPROM.read(25)<<8);
  AcZOff = EEPROM.read(26) + (EEPROM.read(27)<<8);

  // read gyroscope offsets
  GyXOff = EEPROM.read(28) + (EEPROM.read(29)<<8);
  GyYOff = EEPROM.read(30) + (EEPROM.read(31)<<8);
  GyZOff = EEPROM.read(32) + (EEPROM.read(33)<<8);

  // read microservo cal limits
  ServoMin = EEPROM.read(34) + (EEPROM.read(35)<<8);
  ServoMax = EEPROM.read(36) + (EEPROM.read(37)<<8);
  ServoCtr = ServoMin + ((ServoMax - ServoMin)/2);
}

// ---------------------------------------------------------------------------------

void EEPROM_Write() {
  // write valibration data to EEPROM flash memory
  int16_t zSum = 0; byte zLSB = 0; byte zMSB = 0;
  // save IR black thresholds
  zLSB = AvBlk[0]; zSum+= zLSB; zMSB = AvBlk[0]>>8; zSum+= zMSB;
  EEPROM.write( 2,zLSB); EEPROM.write( 3,zMSB); 
  zLSB = AvBlk[1]; zSum+= zLSB; zMSB = AvBlk[1]>>8; zSum+= zMSB;
  EEPROM.write( 4,zLSB); EEPROM.write( 5,zMSB); 
  zLSB = AvBlk[2]; zSum+= zLSB; zMSB = AvBlk[2]>>8; zSum+= zMSB;
  EEPROM.write( 6,zLSB); EEPROM.write( 7,zMSB); 
  zLSB = AvBlk[3]; zSum+= zLSB; zMSB = AvBlk[3]>>8; zSum+= zMSB;
  EEPROM.write( 8,zLSB); EEPROM.write( 9,zMSB); 
  zLSB = AvBlk[4]; zSum+= zLSB; zMSB = AvBlk[4]>>8; zSum+= zMSB;
  EEPROM.write(10,zLSB); EEPROM.write(11,zMSB); 
  
  // save IR white thresholds
  zLSB = AvWht[0]; zSum+= zLSB; zMSB = AvWht[0]>>8; zSum+= zMSB;
  EEPROM.write(12,zLSB); EEPROM.write(13,zMSB); 
  zLSB = AvWht[1]; zSum+= zLSB; zMSB = AvWht[1]>>8; zSum+= zMSB;
  EEPROM.write(14,zLSB); EEPROM.write(15,zMSB); 
  zLSB = AvWht[2]; zSum+= zLSB; zMSB = AvWht[2]>>8; zSum+= zMSB;
  EEPROM.write(16,zLSB); EEPROM.write(17,zMSB); 
  zLSB = AvWht[3]; zSum+= zLSB; zMSB = AvWht[3]>>8; zSum+= zMSB;
  EEPROM.write(18,zLSB); EEPROM.write(19,zMSB); 
  zLSB = AvWht[4]; zSum+= zLSB; zMSB = AvWht[4]>>8; zSum+= zMSB;
  EEPROM.write(20,zLSB); EEPROM.write(21,zMSB);

  // save accelerometer offsets
  zLSB = AcXOff; zSum+= zLSB; zMSB = AcXOff>>8; zSum+= zMSB;
  EEPROM.write(22,zLSB); EEPROM.write(23,zMSB); 
  zLSB = AcYOff; zSum+= zLSB; zMSB = AcYOff>>8; zSum+= zMSB;
  EEPROM.write(24,zLSB); EEPROM.write(25,zMSB); 
  zLSB = AcZOff; zSum+= zLSB; zMSB = AcZOff>>8; zSum+= zMSB;
  EEPROM.write(26,zLSB); EEPROM.write(27,zMSB); 

  // save gyroscope offsets
  zLSB = GyXOff; zSum+= zLSB; zMSB = GyXOff>>8; zSum+= zMSB;
  EEPROM.write(28,zLSB); EEPROM.write(29,zMSB); 
  zLSB = GyYOff; zSum+= zLSB; zMSB = GyYOff>>8; zSum+= zMSB;
  EEPROM.write(30,zLSB); EEPROM.write(31,zMSB); 
  zLSB = GyZOff; zSum+= zLSB; zMSB = GyZOff>>8; zSum+= zMSB;
  EEPROM.write(32,zLSB); EEPROM.write(33,zMSB);

  // save microservo cal limits
  zLSB = ServoMin; zSum+= zLSB; zMSB = ServoMin>>8; zSum+= zMSB;
  EEPROM.write(34,zLSB); EEPROM.write(35,zMSB);
  zLSB = ServoMax; zSum+= zLSB; zMSB = ServoMax>>8; zSum+= zMSB;
  EEPROM.write(36,zLSB); EEPROM.write(37,zMSB);
  
  // write checksum
  // Serial.println("Data Sum =\t "+ String(zSum));
  EEPROM.write(0,zSum); EEPROM.write(1,zSum>>8);
  EEPROM.commit();
}

// ---------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = abs(cmdVal * 10) + zVal;
}

// --------------------------------------------------------------------------------

void flashLEDs(int zCnt) {
  // flash the left and right LED swCnt times whilst running the main loop()
  // note the calling function must block being called from loop() to prevent stack
  // overflow conditions
  LED_Nop = true;  // block eye LED tasks
  for (int zP = 0; zP < zCnt; zP++) {
    fill_solid(pixels,5,CRGB(200, 0, 0)); FastLED.show();
    loopDelay(30);
    FastLED.clear();FastLED.show();
    loopDelay(270);
  }
  LED_Nop = false; // enable eye LED tasks
}

// --------------------------------------------------------------------------------

String GetBIN(byte zVal) {
  // returns an 8 character binary representation of zVal
  String zBIN$ = "";
  // convert zVal into binary
  if (zVal & 0B10000000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B01000000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00100000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00010000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00001000) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000100) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000010) {zBIN$ += "1";} else {zBIN$ += "0";}
  if (zVal & 0B00000001) {zBIN$ += "1";} else {zBIN$ += "0";}
  return zBIN$;
}

// --------------------------------------------------------------------------------

String getHmTime(unsigned long zMillis) {
  // returns a 'time' string in Hrs:Min base on zMillis
  String zTime$ = ""; zMillis = zMillis/60000;  // reduce zMillis to minutes
  long zU = zMillis/60; zMillis = zMillis%60;    // get hours and minutes
  zTime$ = String(zU) + ":";
  if (zMillis < 10) {zTime$ += "0";}
  zTime$ += String(zMillis);  // add minutes
  return zTime$;
}

// --------------------------------------------------------------------------------

String getHmsTime(unsigned long zMillis) {
  // returns a 'time' string in Hrs:Min:sec base on zMillis
  String zTime$ = "";
  long zU = zMillis/3600000; zMillis = zMillis%3600000;
  if (zU > 0) {zTime$ = String(zU) + ":";}
  zU = zMillis/60000; zMillis = zMillis%60000;
  if (zU < 10) {zTime$ += "0";}
  zTime$ += String(zU);
  zU = zMillis/1000;
  if (zU < 10) {zTime$ += ":0";} else {zTime$ += ":";}
  zTime$ += String(zU);
  return zTime$;
}

// -------------------------------------------------------------------------------

String getMaskStr() {
  // returns a 5 character binary mask string, based on the state of the IR_Mask bits
  // the 5-bit IR_Mask sets bits as follows:
  // [ 0 ][ 0 ][ 0 ][RLft][RRht][FRht][FCtr][FLft]
  // but here we change that representation for display purposes to:
  // [RLft][FLft][FCtr][FRht][RRht]
  String zMask$ = "";
  if ((IR_Mask & B10000) > 0) {zMask$ += "1";} else {zMask$ += "0";}  // RLft
  if ((IR_Mask & B00001) > 0) {zMask$ += "1";} else {zMask$ += "0";}  // FLft
  if ((IR_Mask & B00010) > 0) {zMask$ += "1";} else {zMask$ += "0";}  // FCtr
  if ((IR_Mask & B00100) > 0) {zMask$ += "1";} else {zMask$ += "0";}  // FRht
  if ((IR_Mask & B01000) > 0) {zMask$ += "1";} else {zMask$ += "0";}  // RRht
  return zMask$;
}

// -------------------------------------------------------------------------------

String getNumCSV(long zNum ) {
  // returns a string of zNum with commas ie. 1,000,000
  String zVal$;
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000); zNum /= 1000;}
  else {zVal$ = String(zNum); return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  if (zNum > 999) {zVal$ = "," + getNumNNN(zNum % 1000) + zVal$; zNum /= 1000;}
  else {zVal$ = String(zNum) + zVal$; return zVal$;}
  return zVal$;
}

// -------------------------------------------------------------------------------

String getNumNNN(long zNum) {
  // returns a 3 digit string for numbers 000 - 999
  String zVal$ = String(zNum);
  if (zNum < 10) zVal$ = "00" + zVal$;
  else if (zNum < 100) zVal$ = "0" + zVal$;
  return zVal$;
}

// -------------------------------------------------------------------------------

void getRangeCol(int zR) {
  // returns a colour for a range value
  // 0 = OFF
  // 1 = Green
  // 2 = Green/Yellow
  // 3 = Yellow
  // 4 = Orange
  // 5 = Red
  colR = 10; colG = 0; colB = 0;  // default is zR >= 5 Red
  switch (zR) {
    case 0: colR = 0; colG = 0; colB = 0; break;
    case 1: colR = 0; colG = 10; colB = 0; break;
    case 2: colR = 3; colG = 7; colB = 0; break;
    case 3: colR = 5; colG = 5; colB = 0; break;
    case 4: colR = 7; colG = 3; colB = 0; break;
  }
}

// -------------------------------------------------------------------------------

int getServoAng() {
  // return a angle corresponding to servoVal
  // the micro-servo is assumed to swing from 50° to 130°, left to right
  // however servo PWM values increase from right to left
  return map(servoVal,ServoMin,ServoMax,ServoAngMaxDef,ServoAngMinDef);  
}

// --------------------------------------------------------------------------------

void Init_ESP_NOW() {
  // called to initiate an ESEP-NOW link or connect to another trannsceiver
  // each time this function is called it will attempt to connect to a different
  // boradcast address, cycling through them, over and over
//  Serial.println("Init_ESP_NOW()");

  // if we have previously tried a connection then we need to de-initialise it
  if (ESP_NOW_Init) {
    // already attempted to initialise ESP-NOW so strip everything back
    // Serial.println("Closing ESP-NOW");
    esp_now_del_peer(broadcastAddress);
    WiFi.disconnect();
  }
  
  // Init ESP-NOW and returns its status
  if (esp_now_init() != ESP_OK) {
    // Serial.println("Error initializing ESP-NOW");
    return;
  }
  ESP_NOW_Init = true;
  
  // change the broadcast address pointer each time
  ESP_NOW_BA++; if (ESP_NOW_BA > 1) {ESP_NOW_BA = 0;}
  // Serial.println("ESP_NOW_BA = " + String(ESP_NOW_BA));
  
  switch (ESP_NOW_BA) {
    case 0: // Wii Transciever unique MAC: 50:02:91:68:F7:3F
      broadcastAddress[0] = 0x50;
      broadcastAddress[1] = 0x02;
      broadcastAddress[2] = 0x91;
      broadcastAddress[3] = 0x68;
      broadcastAddress[4] = 0xF7;
      broadcastAddress[5] = 0x3F;
      break;
    case 1: // Wii Transciever unique MAC: 50:02:91:68:F0:D2
      broadcastAddress[0] = 0x50;
      broadcastAddress[1] = 0x02;
      broadcastAddress[2] = 0x91;
      broadcastAddress[3] = 0x68;
      broadcastAddress[4] = 0xF0;
      broadcastAddress[5] = 0xD2;
      break;
  }

  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);

  // Register peer
  esp_now_peer_info_t peerInfo; //initialize and assign the peer information as a pointer to an addres
  memset(&peerInfo, 0, sizeof(peerInfo));
  memcpy(peerInfo.peer_addr, broadcastAddress, 6); //copy the value of  broadcastAddress with 6 bytes to peerInfo.peer_addr
  peerInfo.channel = 0;     //channel at which the esp talk. 0 means undefined and data will be sent on the current channel. 1-14 are valid channels which is the same with the local device 
  peerInfo.encrypt = false; //not encrypted

  //Add the device to the paired device list 
  if (esp_now_add_peer(&peerInfo) != ESP_OK){ 
    // Serial.println("Failed to add peer");
    return;
  }

  // Register for a callback function that will be called when data is received
  esp_now_register_recv_cb(OnDataRecv);
}

// -------------------------------------------------------------------------------

void initialiseIR() {
  // called when starting a moving task to determine the mode of the sensor array
  // depending on light/dark surface it will set flags for edge sensing
  // the readSensors() function will then perform the sensing arbitration
  int zAvgRaw = (AvRaw[0] + AvRaw[1] + AvRaw[2] + AvRaw[3] + AvRaw[4])/5; // average sensor value
  int zAvgCal = (AvWht[0] + AvWht[1] + AvWht[2] + AvWht[3] + AvWht[4])/5; // average white threshold
  if (zAvgRaw <= zAvgCal) {IR_BaseF = true;} else {IR_BaseF = false;}
  ADCscan = false;
}

// -------------------------------------------------------------------------------

void initialiseMPU() {
  // By default the MPU-6050 sleeps. So we have to wake it up.
  Wire.beginTransmission(MPU_addr); //Start communication with the address found during search.
  Wire.write(0x6B); //We want to write to the PWR_MGMT_1 register (6B hex)
  Wire.write(0x00); //Set the register bits as 00000000 to activate the gyro
  I2CError = Wire.endTransmission();
  if (I2CError == 0) {
    //Set the full scale of the gyro to +/- 250 degrees per second
    Wire.beginTransmission(MPU_addr); //Start communication with the address found during search.
    Wire.write(0x1B); //We want to write to the GYRO_CONFIG register (1B hex)
    Wire.write(0x00); //Set the register bits as 00000000 (250dps full scale)
    Wire.endTransmission();
    //Set the full scale of the accelerometer to +/- 4g.
    Wire.beginTransmission(MPU_addr); //Start communication with the address found during search.
    Wire.write(0x1C); //We want to write to the ACCEL_CONFIG register (1A hex)
    Wire.write(0x08); //Set the register bits as 00001000 (+/- 4g full scale range)
    Wire.endTransmission();
    //Set some filtering to improve the raw data.
    Wire.beginTransmission(MPU_addr); //Start communication with the address found during search
    Wire.write(0x1A); //We want to write to the CONFIG register (1A hex)
    Wire.write(0x03); //Set the register bits as 00000011 (Set Digital Low Pass Filter to ~43Hz)
    Wire.endTransmission();
    Serial.println("MPU Initialised"); 
  } else {
    Serial.println("MPU Initialisation Failed!"); 
    Serial.println("I2C Error = " + String(I2CError)); 
  }
}

// --------------------------------------------------------------------------------

void invokeCal() {
  // called from reading SW0 & SW1 to invoke a calibration process
  // clear variables and block switch reading
  bool zErr = false;
  sw0Cnt = 0; sw0Timer = 0; sw0DwnTime = 0; ESC0 = false;
  sw1Cnt = 0; sw1Timer = 0; sw1DwnTime = 0; ESC1 = false;
  SW0_Nop = true; SW1_Nop = true; // disable button switch actions
  
  // run the dual button release task before entering calibration mode
  // if sensor calibration task is not available then display ERROR!
  setMainTask(98);  // process task starts once both buttons are released
  if (DispMenu == 0) {
    // we are displaying sensor information
    switch(DispPnt) {
      case 1: // MPU 6050
        Display_Text2x16("Calibrate","MPU Offsets");
        MainNext = 80; break;
      case 2: // LTOF ranger
        Display_Text2x16("Calibrate","LTOF Servo");
        MainNext = 81; break;
      case 4: // IR sensors
        Display_Text2x16("Calibrate","IR Sensors");
        MainNext = 82; break;
      default:
        zErr = true; break;
    }
  } else {zErr = true;}
  if (zErr) {
    Display_Text2x16("ERROR!","Can't Calibrate"); MainNext = 0;
    SW0_Nop = false; SW1_Nop = false; // renable button switch actions
  }
}

// --------------------------------------------------------------------------------

void IR_SensorsOFF() {
  // turns OFF all IR sensors to conserve power  
  digitalWrite(FLftPwr,LOW);  // turn 1st sensor OFF
  digitalWrite(FCtrPwr,LOW);  // turn 2nd sensor OFF
  digitalWrite(FRhtPwr,LOW);  // turn 3rd sensor OFF
  digitalWrite(RRhtPwr,LOW);  // turn 4th sensor OFF
  digitalWrite(RLftPwr,LOW);  // turn 5th sensor OFF
}

// --------------------------------------------------------------------------------

bool IR_TestBlack() {
  // returns true if placed on a black surface, all sensors reading > 1000
  if (AvRaw[0] > 1000) {
    if (AvRaw[1] > 1000) {
      if (AvRaw[2] > 1000) {
        if (AvRaw[3] > 1000) {
          if (AvRaw[4] > 1000) {return true;}
        }
      }
    }
  } return false;
}

// --------------------------------------------------------------------------------

bool IR_TestWhite() {
  // returns true if placed on a white surface, all sensors reading < 1000
  if (AvRaw[0] < 1000) {
    if (AvRaw[1] < 1000) {
      if (AvRaw[2] < 1000) {
        if (AvRaw[3] < 1000) {
          if (AvRaw[4] < 1000) {return true;}
        }
      }
    }
  } return false;
}

// --------------------------------------------------------------------------------

void LED_Tasks() {
  // called from the main loop() every 20ms (50Hz) to perform a range of LED tasks
  // LED_Task functions:
  //  -1  sweep to centre, default
  //   0  do nothing
  // -10  bar graph based on LEDVal
  // -20  red IR sensor lamps based on IR_Mask
  // -30  moving LED based on servo angle
  // -40  dot swing left/right/left
  // -50  attacking signature
  // -60  LED lit == CountDwn sequence
  if (LED_Nop) {return;}

  // check for button press?
  if ((sw0State == LOW) || (sw1State == LOW)) {
    if (LED_SwDwn) {return;}
    Pixel_Clear_Array(); Pixel_SetLED(2, 15, 0, 0); Pixel_ShowArray(false);
    LED_Task = -abs(LED_Task); LED_SwDwn = true; return;
  } else {LED_SwDwn = false;}
  
  // test for the delay skip counter
  if (LED_Del > 0) {LED_Del--; return;}
  
  // when changing task set the LED_Task value -ve to reset variables
  if (LED_Task < 0) {
    // change of LED task mode so clear variables
    LED_Task = abs(LED_Task); LED_SubTask = 0; LED_Cnt = 0; LED_Inc = 1;
    LED_Del = 0; LED_Pnt = -1; LED_IR_Mask = B11111111; LED_Blink = true;
  }

  // perform specified task
  switch(LED_Task) {
    case 0: break;  // NOP
    case 1:
      Pixel_Clear_Array();            // set all LEDs OFF
      LED_Del = 10;                   // set speed timer
      colR = 0; colG = 0; colB = 4;   // set default blue colour
      if (WiFiEn) {colR = 2; colG = 0; colB = 2;}
      if (LTOF_On > 0) {colR = 3; colG = 0; colB = 0;}
      if (GyCal > 0) {colR = 3; colG = 3; colB = 0;}
      if (LTOF_On == 0) {
        switch(LED_SubTask) {
          case 0: Pixel_SetLED(0,colR,colG,colB); Pixel_SetLED(4,colR,colG,colB); break;
          case 1: Pixel_SetLED(1,colR,colG,colB); Pixel_SetLED(3,colR,colG,colB); break;
          case 2: Pixel_SetLED(2,colR,colG,colB); break;
          case 3: LED_Del = 5; break;  // all LEDs off for short period
        }
      } else {
        LED_Del = 5;  // go twice as fast when laser is ON
        switch(LED_SubTask) {
          case 0: Pixel_SetLED(2,colR,colG,colB); break;
          case 1: Pixel_SetLED(1,colR,colG,colB); Pixel_SetLED(3,colR,colG,colB); break;
          case 2: Pixel_SetLED(0,colR,colG,colB); Pixel_SetLED(4,colR,colG,colB); break;
          case 3: LED_Del = 2; break;  // all LEDs off for short period
        }
      }
      LED_SubTask++; if (LED_SubTask > 3) {LED_SubTask = 0;}
      Pixel_ShowArray(false); break;
    
    case 10: // bar graph task
      // LEDVal 0 - 5 determines which LEDs are lit
      getRangeCol(LEDVal);
      for (int zP = 0; zP < 5; zP++) {
        if (LEDVal > zP) {Pixel_SetLED(zP,colR,colG,colB);}
        else {Pixel_SetLED(zP, 0, 0, 0);}
      } break; 

    case 20:
      //red IR sensor lamps based on IR_Mask
      if (IR_Mask != LED_IR_Mask) {
        // LED mask has changed, so change them all to suit
        // LED strip is written to from right to left
        // [ 0 ][ 0 ][ 0 ][RLft][RRht][FRht][FCtr][FLft] - mask bits
        //                [  4 ][  0 ][  1 ][  2 ][  3 ] - LEDs in strip   
        if ((IR_Mask & 1) > 0) {Pixel_SetLED(3,30,0,0);} else {Pixel_SetLED(3,1,0,0);}
        if ((IR_Mask & 2) > 0) {Pixel_SetLED(2,30,0,0);} else {Pixel_SetLED(2,1,0,0);}
        if ((IR_Mask & 4) > 0) {Pixel_SetLED(1,30,0,0);} else {Pixel_SetLED(1,1,0,0);}
        if ((IR_Mask & 8) > 0) {Pixel_SetLED(0,30,0,0);} else {Pixel_SetLED(0,1,0,0);}
        if ((IR_Mask & 16) > 0) {Pixel_SetLED(4,30,0,0);} else {Pixel_SetLED(4,1,0,0);}
        LED_IR_Mask = IR_Mask;
      } break;

    case 30:
      //red LTOF sensor angle based on microservo
      {
      int16_t zLED_Pnt = LED_Pnt;
      int16_t zTgt;
           if (ServoAng >= 114) {LED_Pnt = 0;}
      else if (ServoAng >=  98) {LED_Pnt = 1;}
      else if (ServoAng >=  82) {LED_Pnt = 2;}
      else if (ServoAng >=  66) {LED_Pnt = 3;}
      else {LED_Pnt = 4;}
      if (RangeFtdTgt < 500) {
        // target within range so get its LED value
             if (ServoAngTgtAvg >= 114) {zTgt = 0;}
        else if (ServoAngTgtAvg >=  98) {zTgt = 1;}
        else if (ServoAngTgtAvg >=  82) {zTgt = 2;}
        else if (ServoAngTgtAvg >=  66) {zTgt = 3;}
        else {zTgt = 4;}
      } else {zTgt = -1;} // target not within range
      if (zTgt == LED_Pnt) {LED_Pnt+= 5;}
      if (LED_Pnt != zLED_Pnt) {
        // pointer has changed
        Pixel_SetAllLEDs(1,0,0);         // set all LEDs low RED
        if (LED_Pnt > 4) {LED_Pnt-= 5; Pixel_SetLED(LED_Pnt,0,0,40);}
        else {Pixel_SetLED(LED_Pnt,30,0,0);}
        Pixel_ShowArray(false);
      }} break;

    case 40:
      // swinging dot left/right/left
      Pixel_SetAllLEDs(0,0,0);          // set all LEDs OFF
      Pixel_SetLED(LED_Cnt,30,30,0);    // set pointer LED to yellow
      LED_Cnt += LED_Inc;
           if (LED_Cnt > 4) {LED_Cnt = 3; LED_Inc = -1;}
      else if (LED_Cnt < 0) {LED_Cnt = 1; LED_Inc = 1;}
      LED_Del = 1; break;

    case 50:
      // attacking signature, bright red, centre blue, flashing
      if (LED_Blink) {
        Pixel_SetLED(0,20,0,0);
        Pixel_SetLED(1,0,0,40);
        Pixel_SetLED(2,0,0,60);
        Pixel_SetLED(3,0,0,40);
        Pixel_SetLED(4,20,0,0);
      } else {
        Pixel_SetLED(0,1,0,0);
        Pixel_SetLED(1,0,0,2);
        Pixel_SetLED(2,0,0,4);
        Pixel_SetLED(3,0,0,3);
        Pixel_SetLED(4,1,0,0);
      }
      Pixel_ShowArray(false);
      LED_Del = 10; break;

    case 60:
      // LED lit == CountDown sequence
      // once set the LED task function goes into NOP mode
      Pixel_SetAllLEDs(0,0,0); // clear all LEDs
      Pixel_SetLED(CountDown,30,30,0);
      LED_Task = 0; Pixel_ShowArray(false); break;
  }
  LED_Blink = !LED_Blink; // toggle the 'blink' flag
}

// --------------------------------------------------------------------------------

void LTOF_Read_Tasks() {
  // called from the main loop when enabled to read the laser range finder
  switch(LTOF_On) {
    // this is the VL53L0X_readRangeContinuousMillimeters() function broken down
    // into a state machine which take measurements when LTOF_On > 0
    // use the VL53L0X_Enable/Disable functions to switch the laser ON/OFF
    case 0: break;
    case 1: 
      // Returns a range reading in millimeters when continuous mode is active
      // (readRangeSingleMillimeters() also calls this function after starting a
      // single-shot range measurement)
      startTimeout(); LTOF_On++; break;
    case 2:
      // This is the main loop where we are waiting for the measurement to be completed
      // so we can branch to a multitasking loop from here whilst waiting
      // note that readReg takes about 500 us so not a fast while loop
      if ((VL53L0X_readReg(RESULT_INTERRUPT_STATUS) & 0x07) == 0) {
        // if no interrupt flag then check to see if a timeout has occured
        if (checkTimeoutExpired()) {
          // a timeout has occured
          did_timeout = true;
//          return 65535;
          Serial.println(F("LTOF timeout error!"));
        }
      } else {
        // an interrupt has occured suggesting a valid measurement
//        t1 = millis(); Serial.println(t1 - t0); t0 = t1; 
       LTOF_Wait = millis() + 32;  // default is +28
       LTOF_On++;
      } break;
    case 3:
      // assumptions: Linearity Corrective Gain is 1000 (default);
      // fractional ranging is not enabled
      Range = VL53L0X_readReg16Bit(RESULT_RANGE_STATUS + 10);
      Status = (VL53L0X_readReg(RESULT_RANGE_STATUS) & 0x78) >> 3;
      RangeNew = true;  // flag availability of new range, this must be cleared by monitoring code
      VL53L0X_writeReg(SYSTEM_INTERRUPT_CLEAR, 0x01);
      LTOF_t0 = LTOF_t1; LTOF_t1 = millis();  // set timers for fps
      LTOF_fps = 1000/(LTOF_t1 - LTOF_t0);
      // limit raw Range to 1200, ie 1.2m as it gets very noisy beyond this
      if (Range > RangeOL) {Range = RangeOL;}
//      Serial.println(Range);
      LTOF_On++; break;
    case 4:
      // apply filtering to raw Range measurement
      RangeFtd = Range;
      if (Status != 11) {RangeFtd = 999;}
      if (RangeFtd >= 999) {
        RangeFtd = 999;
        if ((RangeFtd - (int)RangeLast) > 200) {
          Pulse++;
          if (Pulse < 3) {RangeFtd = RangeLast;}
        } else {Pulse = 0;}
      } else {
        if (((int)RangeLast - RangeFtd) > 200) {
          Pulse++;
          if (Pulse < 3) {RangeFtd = RangeLast;}
        } else {Pulse = 0;}
        if (RangeFtd < RangeMin) {RangeFtd = RangeMin;} // set minimum distance
      }
      // apply a filtered range limit of RangeMax
      RangeLast = RangeFtd; RangeFtd = min(RangeFtd,RangeMax);
//      Serial.print(min((int)Range,1000)); Serial.print(","); Serial.println(RangeFtd);
      RangeFtdNew = true; LTOF_On++; break;
    case 5:
      // as the sensor only returns measurements every 30-32ms, there is no point in
      // spending time reading interrupt registers
      if (millis() >= LTOF_Wait) {LTOF_On = 1;}
      break;
  }
}

// --------------------------------------------------------------------------------

void Pixel_Clear_Array() {
  // clears the contents of the colour array
  for (int zL = 0; zL < 5; zL++) {
    colBlu[zL] = 0; colGrn[zL] = 0; colRed[zL] = 0;
  }
}

// --------------------------------------------------------------------------------

void Pixel_RotateLft() {
  // rotate the contents of the colour array  
  colB = colBlu[0]; colG = colGrn[0]; colR = colRed[0];
  for (int zP = 0; zP < 4; zP++) {
    colBlu[zP] = colBlu[zP + 1]; colGrn[zP] = colGrn[zP + 1]; colRed[zP] = colRed[zP + 1];
  }
  colBlu[4] = colB; colGrn[4] = colG; colRed[4] = colR;
}

// --------------------------------------------------------------------------------

void Pixel_RotateRht() {
  // rotate the contents of the colour array  
  colB = colBlu[4]; colG = colGrn[4]; colR = colRed[4];
  for (int zP = 4; zP > 0; zP--) {
    colBlu[zP] = colBlu[zP -1]; colGrn[zP] = colGrn[zP -1]; colRed[zP] = colRed[zP -1];
  }
  colBlu[0] = colB; colGrn[0] = colG; colRed[0] = colR;
}

// --------------------------------------------------------------------------------

void Pixel_SetAllLEDs(byte zR, byte zG, byte zB) {
  // set all LEDs to one colour
  for (int zP = 0; zP < 5; zP++) {
    colBlu[zP] = zB; colGrn[zP] = zG; colRed[zP] = zR;
    pixels[zP].setRGB(colRed[zP],colGrn[zP],colBlu[zP]);
  } LEDshow = true;
}

// --------------------------------------------------------------------------------

void Pixel_SetLED(byte zPixNum, byte zR, byte zG, byte zB) {
  // set the colour of LED zPixNum.
  colBlu[zPixNum] = zB; colGrn[zPixNum] = zG; colRed[zPixNum] = zR;
  pixels[zPixNum].setRGB(colRed[zPixNum],colGrn[zPixNum],colBlu[zPixNum]);
  LEDshow = true;
}

// --------------------------------------------------------------------------------

void Pixel_ShowArray(bool zShow) {
  // load all LED colours and either show them immediately or set a flag
  for (int zP = 0; zP < 5; zP++) {
    pixels[zP].setRGB(colRed[zP],colGrn[zP],colBlu[zP]);
  }
  if (zShow) {FastLED.show(); LEDshow = false;} else {LEDshow = true;}
}

// -------------------------------------------------------------------------------

void readAccelAll() {
  // read the X,Y,Z-axis acceleration values from the MPU
  // in this robot:
  // AcX = yaw, +ve tip right, -ve tip left
  // AcY = up/down, +ve upside-down, -ve right way up
  // AcZ = pitch, +ve tip forward, -ve tip backward
//  t0 = micros();  // t1-t0 average = 225µs, longer with I2C bus contention
  AcXL = AcX; AcYL = AcY; AcZL = AcZ; // track previous values
  Wire.beginTransmission(MPU_addr);
  Wire.write(0x3B);  // starting with register 0x3B (ACCEL_XOUT_H)
  Wire.endTransmission(false);
  Wire.requestFrom(MPU_addr,6,true);  // request 1 register read
  AcX = Wire.read()<<8|Wire.read();  // 0x3B (ACCEL_XOUT_H) & 0x3C (ACCEL_XOUT_L)
  AcY = Wire.read()<<8|Wire.read();  // 0x3D (ACCEL_YOUT_H) & 0x3E (ACCEL_YOUT_L)
  AcZ = Wire.read()<<8|Wire.read();  // 0x3F (ACCEL_ZOUT_H) & 0x40 (ACCEL_ZOUT_L)
  
  // apply calibration offsets to these raw values
  AcX += AcXOff;  // note AcXOff should be zero initially in order to determin its value
  AcY += AcYOff;  // note AcYOff should be zero initially in order to determin its value
  AcZ += AcZOff;  // note AcZOff should be zero initially in order to determin its value

  // if 1st reading then we don't have a valid previous value, so use current
  if (AcAvgDiv < 1) {AcXL = AcX; AcYL = AcY; AcZL = AcZ;}

  // adjust for Pitch offset error if not in calibration mode
  if (!MPUCal) {
         if (AcZ > 0) {AcZ = map(AcZ,0,8192 + AcZOff,0,8192);}
    else if (AcZ < 0) {AcZ = map(AcZ,-8192 + AcZOff,0,-8192,0);}
  }

  // limit values to avoid trig function problems
  if (AcY > 8192) {AcY = 8192;} else if (AcY < -8192) {AcY = -8192;}
  if (AcX > 8192) {AcX = 8192;} else if (AcX < -8192) {AcX = -8192;}
  if (AcZ > 8192) {AcZ = 8192;} else if (AcZ < -8192) {AcZ = -8192;}

  // average over two readings to reduce random noise
  AcX = (AcX + AcXL)/2; AcY = (AcY + AcYL)/2; AcZ = (AcZ + AcZL)/2;

  if (AcY <= 0) {UpRht = true;} else {UpRht = false;} // set the upright flag
//  t1 = micros();
//    Serial.print(AcX); Serial.print("\t");
//    Serial.print(AcY); Serial.print("\t");
//    Serial.println(AcZ);
}

// -------------------------------------------------------------------------------

//void readAccelX() {
//  // read the X-axis acceleration value from the MPU
//  Wire.beginTransmission(MPU_addr);
//  Wire.write(0x3B);  // starting with register 0x3B (ACCEL_XOUT_H)
//  Wire.endTransmission(false);
//  Wire.requestFrom(MPU_addr,2,true);  // request 1 register read
//  AcX = Wire.read()<<8|Wire.read();  // 0x3B (ACCEL_XOUT_H) & 0x3C (ACCEL_XOUT_L)
//  PrintTx += "AcX Raw = " + String(AcX) + "\n";
//}

// -------------------------------------------------------------------------------

//void readAccelY() {
//  // read the Y-axis acceleration value from the MPU
//  // AcY = 8,192 at 1g. In linerar region divide by 138.5 to get degrees
//  Wire.beginTransmission(MPU_addr);
//  Wire.write(0x3D);  // starting with register 0x3D (ACCEL_YOUT_H)
//  I2CError = Wire.endTransmission(false);
//  if (I2CError == 0) {
//    Wire.requestFrom(MPU_addr,2,true);  // request 1 register read
//    if (Wire.available() > 1) {
//      AcY = Wire.read()<<8|Wire.read();  // 0x3D (ACCEL_YOUT_H) & 0x3E (ACCEL_YOUT_L)
//      PrintTx += "AcY Raw = " + String(AcY) + "\n";
//    }
//  }
//  // we apply an offset if needed, depending on how it has been mounted
//  // uncomment the next line to determine the raw value at rest, AcYrest
//  // then set AcYOff = -AcYrest to zero out the offset
//  return;
//  AcYsum += (long)AcY; AcYsum -= AcYaV;   // update the rolling accumulator
//  AcYaV = AcYsum/4;                       // calc the average value
////  if (Fn) {AcYaV = AcYsum/40;}            // greater average in special function mode
////  AcYaV = AcY;                          // don't calc the average value
//    PrintTx += "AcY Av = " + String(AcYaV) + "\n";
////  AcYF = ((float)((AcYaV + AcYOff)/69))/2.0;         // convert to an approximate angle
////  AcYsgn = true; // set sign flag as positive
////  if (AcYF < 0.0){AcYF = -AcYF; AcYsgn = false;}; // set sign flag as negative
//}

// -------------------------------------------------------------------------------

//void readAccelZ() {
//  // read the Z-axis acceleration value from the MPU
//  Wire.beginTransmission(MPU_addr);
//  Wire.write(0x3F);  // starting with register 0x3F (ACCEL_ZOUT_H)
//  Wire.endTransmission(false);
//  Wire.requestFrom(MPU_addr,2,true);  // request 1 register read
//  AcZ=Wire.read()<<8|Wire.read();  // 0x3F (ACCEL_ZOUT_H) & 0x40 (ACCEL_ZOUT_L)
//  AcZ=AcZ + AcZOff;
//}

// --------------------------------------------------------------------------------

void readBattery(int zM) {
  // called from the readSensors() function every 200ms
  // read the voltage on the analog input and average it over ten counts
  // zM = 0 - simple read + average
  // zM = 1 - calc float and percentage
  BatVol = analogRead(BatPin); BatTms = millis(); // take the reading
  BattNow = false;  // clear the battery read flag
//  Serial.println("BatVol = " + String(BatVol));
  BatSum = BatSum + BatVol - BatAvg;
  BatAvg = BatSum/20;     // ADC is averaged over 20 readings to remove noise
  // uncomment the following line if you want to see readings on serial monitor
//  Serial.print(F("BattAv = ")); Serial.print(BattAv); Serial.print(F("\t")); Serial.print((5.0 * BattAv)/608.9); Serial.println(F("v"));
  
  //#############################################################################
  //
  // Battery life prediction code + data logger
  //
  //#############################################################################
  // BatDel is set to 10 second interval to start with, to ignore initial readings
  if (!USBPwr) {
    if (BatDel > 0) {BatDel--;}
    else {
      BatDel = 5;  // use a 1 second timer for prediction updates & data logging
      if (BatAvg > BatAvgMax) {BatAvgMax = BatAvg;}
      BatData[BatDP] = (62L * (BatAvg - BatCritical))/(BatAvgMax - BatCritical);
//      Serial.println("BatData: " + String(BatData[BatDP]));
      if (BatDP < 59) {BatDP++;} else {shiftBatData();}
      
      if (BatT0 == 0) {
        // this is the first measurement after 10 second start delay
        // this values are the cornerstone of the prediction
        BatT0 = BatTms; BatV0 = BatAvg; BatV1 = BatAvg;
//        Serial.print("BatT0 "); Serial.println(BatT0);
      } else {
        if (BatAvg != BatV1) {
          // there has been a change in the battery voltage setting a new discharge point
          // and the current battery voltage is different, so avoid divide by zero error
          // recalculate battery life predicition base on new value and time
//          Serial.print("BatAvg "); Serial.println(BatAvg);
          BatV1 = BatAvg; // record the latest low point
          if  (BatV0 == BatV1) {BatV1--;} // avoid divide by zero error
          AnyLong = ((BatTms - BatT0)*(BatV1 - BatCritical))/(BatV0 - BatV1);
      //      Serial.print("AnyLong "); Serial.println(AnyLong);
          if (AnyLong > 0) {
            BatTime$ = getHmTime(AnyLong);
          }
        }
      }
    }  
  
  //##############################################################################
  //
  //  critical battery test
  //
  //##############################################################################
  // test for battery low condition, 683 == 6.6v
    if (BatAvg < BatCritical) {
      // battery has reached critical level so squat down and blink centre LED
      BatCritCnt++;
      if (BatCritCnt >= 100) {
        // low voltage sustained for 4 seconds or more, so we shut down
        // the delay ignores brown-outs caused by heavy servo movements
        FastLED.clear(); FastLED.show();          // turn OFF all LEDs
        Pixel_SetLED(2, 1, 0, 0); FastLED.show(); // centre LED on dim red
        Display_Text2x16("LOW","BATTERY");
        MoveSTOP(0); ServoOFF(); IR_SensorsOFF();
        delay(3000);
        display.clear(); display.display();
        
        while (true) {
          // blink mouth centre LED red forever
          delay(20); BatCnt--;
          if (BatCnt < 1) {
            BatCnt = 100; Pixel_SetLED(2, 0, 0, 0); FastLED.show();
          }
          if (BatCnt == 5) {Pixel_SetLED(2, 1, 0, 0); FastLED.show();}
          // continue to take readings in case this is a deliberate power-down
          BatVol = analogRead(BatPin);
          BatSum = BatSum + BatVol - BatAvg; BatAvg = BatSum /20;
          // test for user button press
          if ((digitalRead(sw0Pin) == LOW) && (BatAvg > (BatCritical + 10))) {break;}
          yield();
        }
        // user has pressed button during power-down
        // battery looks ok so revive system
        setDefaults();
        setMainTask(0);
        synchLoopTimers(0); // restart loop timers
      }
    } else {BatCritCnt = 0;}  // reset the 4 second timer
  }
  
//  if (Ping == 9) {Ping = 0; zM = 1;} // force a battery report every 10 pings
  if ((zM < 1) && (millis() < BatRep)) {return;} // exit if not a report measurement
  
  // send battery voltage and % to serial port
  BatRep = millis() + 10000; // update reading every 10 seconds
  
  AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
  if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
  if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
//  Serial.print("Batt " + String(float(BatAvg)/BatCal) + "v " + String(AnyLong) + "%\n");
}

// -------------------------------------------------------------------------------

void readGyrosAll() {
  // read the X,Y,Z-axis gyroscope values from the MPU
  // in this robot:
  // GyX = yaw
  // GyY = rotate round
  // GyZ = pitch
  // note gyro values are in degrees/sec so we use time between measurements to 
  // determine the average displaced angle
//  t0 = micros();  // t1-t0 average = 315µs, longer with I2C bus contention
  GyXL = GyX; GyYL = GyY; GyZL = GyZ; // track previous values
  Wire.beginTransmission(MPU_addr);
  Wire.write(0x43);  // starting with register 0x43 (GYRO_XOUT_H)
  Wire.endTransmission(false);
  Wire.requestFrom(MPU_addr,6,true);  // request 1 register read
  GyX = Wire.read()<<8|Wire.read();  // 0x43 (GYRO_XOUT_H) & 0x44 (GYRO_XOUT_L)
  GyY = Wire.read()<<8|Wire.read();  // 0x45 (GYRO_YOUT_H) & 0x46 (GYRO_YOUT_L)
  GyZ = Wire.read()<<8|Wire.read();  // 0x47 (GYRO_ZOUT_H) & 0x48 (GYRO_ZOUT_L)
  if (GyT > 0) {
    GyTD = millis() - GyT;  // get the differential time between readings
    GyT = GyTD + GyT;       // set the timer for the next reading
  } else {
    // 1st reading after reset
    GyT = millis();
    GyTD = 40; GyXL = GyX; GyYL = GyY; GyZL = GyZ;
  }

  // apply calibration offsets to these raw values
  GyX += GyXOff;  // note GyXOff should be zero initially in order to determin its value
  GyY += GyYOff;  // note GyYOff should be zero initially in order to determin its value
  GyZ += GyZOff;  // note GyZOff should be zero initially in order to determin its value

  // average over two readings to reduce random noise
  GyX = (GyX + GyXL)/2; GyY = (GyY + GyYL)/2; GyZ = (GyZ + GyZL)/2;
//  t1 = micros();
}

// --------------------------------------------------------------------------------

void readSensors() {
  // called every 5ms from the main loop to read the analogue sensor inputs
  // the battery is read every 200ms (5Hz) irrespective of the IR sensor scanning
  BatRead--; if (BatRead < 1) {BatRead = 40; BattNow = true;}

  // when in ADC_Fast mode the IR sensors are read at 5ms intervals (40Hz)
  // otherwise this process is slowed right down to 200ms intervals (5Hz)
  ADC_Cnt--; if (ADC_Cnt > 0) {
    if (BattNow) {readBattery(0);}
    return;
  }
  if (ADC_Fast) {ADC_Cnt = 0;} else {ADC_Cnt = 40;}
  int zAv;
  // we read them in a clockwise direction, starting front left
  switch(ADC) {
    case 0: // read 1st sensor then switch to the next
      AvRaw[ADC] = analogRead(FLftACh); // read front left ADC
      digitalWrite(FLftPwr,LOW);        // turn 1st sensor OFF
      if (BattNow) {readBattery(0);}    // interspersed battery readings
      digitalWrite(FCtrPwr,HIGH);       // turn 2nd sensor ON
      zAv = analogRead(FCtrACh);        // read front centre ADC, this will be zero
//        Serial.print(String(Av0) + "\t");
      break;
    case 1: // read 2nd sensor then switch to the next
      AvRaw[ADC] = analogRead(FCtrACh); // read front centre ADC
      digitalWrite(FCtrPwr,LOW);        // turn 2nd sensor OFF
      if (BattNow) {readBattery(0);}    // interspersed battery readings
      digitalWrite(FRhtPwr,HIGH);       // turn 3rd sensor ON
      zAv = analogRead(FRhtACh);        // read front right ADC, this will be zero
//        Serial.print(String(Av1) + "\t");
      break;
    case 2: // read 3rd sensor then switch to the next
      AvRaw[ADC] = analogRead(FRhtACh); // read front right ADC
      digitalWrite(FRhtPwr,LOW);        // turn 3rd sensor OFF
      if (BattNow) {readBattery(0);}    // interspersed battery readings
      digitalWrite(RRhtPwr,HIGH);       // turn 4th sensor ON
      zAv = analogRead(RRhtACh);        // read rear right ADC, this will be zero
//        Serial.print(String(Av2) + "\t");
      break;
    case 3: // read 4th sensor then switch to the next
      AvRaw[ADC] = analogRead(RRhtACh); // read rear right ADC
      digitalWrite(RRhtPwr,LOW);        // turn 4th sensor OFF
      if (BattNow) {readBattery(0);}    // interspersed battery readings
      digitalWrite(RLftPwr,HIGH);       // turn 5th sensor ON
      zAv = analogRead(RLftACh);        // read rear left ADC, this will be zero
//        Serial.print(String(Av3) + "\t");
      break;
    case 4: // read 5th sensor then switch to the 1st
      AvRaw[ADC] = analogRead(RLftACh); // read rear left ADC
      digitalWrite(RLftPwr,LOW);        // turn 5th sensor OFF
      if (BattNow) {readBattery(0);}    // interspersed battery readings
      digitalWrite(FLftPwr,HIGH);       // turn 1st sensor ON
      zAv = analogRead(FLftACh);        // read front left ADC, this will be zero
//        Serial.print(String(Av4) + "\n");
      break;
  }
  // convert reading to a percentage, depending on the base colour, white or black
  if (IR_BaseF) {Av100[ADC] = map(AvRaw[ADC],AvWht[ADC],AvBlk[ADC],0,100);} // white base colour
  else {Av100[ADC] = map(AvRaw[ADC],AvBlk[ADC],AvWht[ADC],0,100);}          // black base colour
  Av100[ADC] = constrain(Av100[ADC],0,100);
  // create a binary mask representing the 5 sensors
  // since the floor will generally be dull or non-reflective we set a threshold of 25%
  // [ 0 ][ 0 ][ 0 ][RLft][RRht][FRht][FCtr][FLft]
  switch(ADC) {
    case 0: if (Av100[0] > 25) {IR_Mask |= B00000001;} else {IR_Mask &= B00011110;} break;
    case 1: if (Av100[1] > 25) {IR_Mask |= B00000010;} else {IR_Mask &= B00011101;} break;
    case 2: if (Av100[2] > 25) {IR_Mask |= B00000100;} else {IR_Mask &= B00011011;} break;
    case 3: if (Av100[3] > 25) {IR_Mask |= B00001000;} else {IR_Mask &= B00010111;} break;
    case 4: if (Av100[4] > 25) {IR_Mask |= B00010000;} else {IR_Mask &= B00001111;} break;
  }
  
  // Data recorder
  if (ADC == ADC_Rec) {
    // record this channel
    ADC_RecPnt++; if (ADC_RecPnt > 39) {ADC_RecPnt = 0;}
    ADC_Data[ADC_RecPnt] = AvRaw[ADC];
  }
  ADC++; if (ADC > 4) {ADC = 0; ADCscan = true;}
}

// --------------------------------------------------------------------------------

void readSerial() {
  // reads characters from the serial port and responds to commands
  // set PrintTgt to indicate source as serial port for responses
  keyVal = Serial.read();
  if (keyVal != -1) {PrintTgt = 0; decodeKey(keyVal);}
}

// --------------------------------------------------------------------------------

void readSW0() {
  // called every 10ms from the main loop
  // reads the left button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  sw0State = digitalRead(sw0Pin) & DMsw0; // record button state and display monitor actions
  if (SW0_Nop) {return;}                  // block this function whilst == true
  
  bool sw0Clr = false;
  if (sw0State == LOW) {
    //##############################################################################
    //
    // SW0 button is pressed down
    //
    //##############################################################################
    if (MainTask > 0) {waitMainEnd(); return;}
    if (sw0LastState == HIGH) {
      // SW0 has just been pressed down
  //    Serial.println("SW0 Hi-LO");
      sw0DwnTime = 0; // restart button down time counter
      sw0Cnt++; // count on the falling edge
      sw0Timer = 0; // reset the timer
    } else {
      sw0DwnTime++; // track button pressed time
    }
    if ((sw0DwnTime >= 60) && (!DispNOP)) {
      // long press >= 600ms so present 'BACK' arrow if in Disp[Menu == 0
      if ((!DispBack) && (DispMenu == 0)) {DispCnt = 0; DispBack = true; Display_40ms();}  // force display update with a back arrow
    }
    if ((sw0DwnTime >= 120) && (!DispNOP)) {
      // SW0 held down for >= 1.2 second so toggle menu modes if released before 3 seconds
      DispBack = false;
      if (DispMenu != 2) {
        Display_Overlay_Sumo(); DispMenu = 2;     // display SUMO mode
      } else {
        Display_Overlay_Sensors(); DispMenu = 0;  // display Battery sensor
      }
      ADC_Fast = false; ServoTaskPnt = -1; VL53L0X_OFF ();
      DispNOP = true; DispClr = true;
    }
    if (sw1State == LOW) {invokeCal(); return;} // both buttons pressed!
  } else {
    //##############################################################################
    //
    // SW0 button is released
    //
    //##############################################################################
    if (sw0LastState == LOW) {
      // SW0 has just been released
  //    Serial.println("SW0 Lo-Hi");
      DispBack = false;
      if (DispNOP) {
        // we are changing the display menu mode
        DispNOP = false; DispClr = true; DispDisp = false; DispCnt = 0;
        DispPnt = 0; DispMode = 0;
        sw0Cnt = 0; sw0Timer = 0;
      }
    }
    if (sw0Timer > 0) {
      if (sw0DwnTime < 60) {
        // short press so check what mode the display is in?
        // if not in an action mode simply switch the display mode
        LED_Task = -1;    // set LED tasks to default unless changed here
        DispChng = true;  // flag to initialise certain functions
        switch(DispMenu) {
          case 0: // in 'Sensors' display mode, so toggle between sensors
            switch(DispPnt) {
              case 0: // battery display
                DispMode++; if (DispMode > 1) {DispMode = 0;}
                break;
              case 1: // MPU display
                DispMode++; if (DispMode > 4) {DispMode = 0;}
                break;
              case 2: // LTOF display
                DispMode++; if (DispMode > 4) {DispMode = 0;}
                break;
              case 3: // WiFi display
                DispMode++; if (DispMode > 3) {DispMode = 0;}
                break;
              case 4: // IR Sensor display
                DispMode++; if (DispMode > 7) {DispMode = 0;}
                initialiseIR(); LED_Task = -20;
                if (DispMode > 2) {clearADCdata();}
                break;
            } break;
          case 1: // in 'Tasks' display mode, so initiate tasks
            switch(DispPnt) {
              case 0: // Drive Task
                setMainTask(2); break;
              case 1: // Hunt Task
                setMainTask(3); break;
              case 2: // Tip Test
                setMainTask(4); break;
              case 3: // Target Test
                setMainTask(5); break;
              case 4: // Distance Test
                setMainTask(6); break;
            } break;
         case 2: // in 'Sumo' display mode, so initiate games
            switch(DispPnt) {
              case 0: // Sumo scout and engage Task
                setMainTask(10); break;
            } break;
        }
      } else {
        // button pressed longer so decrement screen displayed
        LED_Task = -1;    // set LED tasks to default unless changed here
        DispChng = true;  // flag to initialise certain functions
        if (DispMenu == 0) {
          // in sensor menu mode
          switch(DispPnt) {
            case 0: // battery display
              DispMode--; if (DispMode < 0) {DispMode = 1;}
              break;
            case 1: // MPU display
              DispMode--; if (DispMode < 0) {DispMode = 4;}
              break;
            case 2: // LTOF display
              DispMode--; if (DispMode < 0) {DispMode = 4;}
            break;
            case 3: // WiFi display
              DispMode--; if (DispMode < 0) {DispMode = 1;}
              break;
            case 4: // IR Sensor display
              DispMode--; if (DispMode < 0) {DispMode = 7;}
              break;
          }
        }
      }
        // button released for 1 sec so assume valid button count
    //    Serial.print(F("swCnt = ")); Serial.println(swCnt);
//        SW0_Nop = true; flashLEDs(sw0Cnt); // indicate count by flashing
//        SW0_Nop = false;
      sw0Cnt = 0; sw0Timer = 0; sw0DwnTime = 0; DispCnt = 0;
    }
  }
  if (ESC0 && !sw0State) {return;}
  sw0LastState = sw0State; // record current state
  ESC0 = false;
  if (sw0Cnt > 0) {sw0Timer++;}
}

// --------------------------------------------------------------------------------

void readSW1() {
  // called every 10ms from the main loop
  // read the right button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  sw1State = digitalRead(sw1Pin) & DMsw1; // record button state and display monitor actions
  if (SW1_Nop) {return;}          // block this function whilst == true
  
  if (sw1State == LOW) {
    //##############################################################################
    //
    // SW1 button is pressed down
    //
    //##############################################################################
    if (MainTask > 0) {waitMainEnd(); return;}
    if (DispCnt < 10) {DispCnt = 10;} // don't allow display to change when button is pressed
    if (sw1LastState == HIGH) {
      // SW1 has just been pressed down
//      Serial.println("SW1 Hi-LO");
      sw1DwnTime = 0; // restart button down time counter
      sw1Cnt++; // count on the falling edge
      sw1Timer = 0; // reset the timer
//    if (MainTask) {sw1Cnt = 0; SetMainTask(98); ESC1 = true;}
    } else {
      sw1DwnTime++; // track button pressed time
    }
    if (ESC1 && !sw1State) {return;}
    ESC1 = false;
    if ((sw1DwnTime >= 60) && (!DispNOP)) {
      // long press so present 'BACK' arrow
      if (!DispBack) {DispCnt = 0; DispBack = true; Display_40ms();}  // force display update with a back arrow
    }
    if ((sw1DwnTime >= 120) && (!DispNOP)) {
      // SW1 held down for >= 1 second so toggle menu modes if released before 3 seconds
      DispBack = false;
      if (DispMenu != 1) {
        Display_Overlay_Tasks(); DispMenu = 1;    // display TASKS screen
      } else {
        Display_Overlay_Sensors(); DispMenu = 0;  // display Battery screen
      }
      ADC_Fast = false; ServoTaskPnt = -1; VL53L0X_OFF ();
      DispNOP = true; DispClr = true;
    }
  } else {
    //##############################################################################
    //
    // sw1State == HIGH as SW1 button is not pressed
    //
    //##############################################################################
    if (sw1LastState == LOW) {
      // SW1 has just been released
//      Serial.println("SW1 Lo-Hi");
      DispBack = false;
      if (DispNOP) {
        // we are changing the display menu mode
        DispNOP = false; DispClr = true; DispDisp = false; DispCnt = 0;
        DispPnt = 0; DispMode = 0;
        sw1Cnt = 0; sw1Timer = 0;
      }
    }
    if (sw1Timer > 0) {
//        PrintTx += "sw1Timer = " + String(sw1Timer) + "\n";
      if (sw1DwnTime < 60) {
        // short press so increment the display type pointer
        LED_Task = -1;    // set LED tasks to default unless changed here
        DispChng = true;  // flag to initialise certain functions
        switch (DispMenu) {
          case 0: // in 'Sensors' display mode, so toggle sensor sub-screens
            DispPnt++; if (DispPnt > 4) {DispPnt = 0;}
            if (DispPnt == 4) { // turn ON fast read mode
              ADC_Fast = true;
              initialiseIR(); LED_Task = -20;
            } else {
              ADC_Fast = false; LED_Task = -1;
            }
            DispMode = 0; // reset the display mode
            if (LTOF_On > 0) {VL53L0X_OFF();} // LTOF ON so switch it off
            if (ServoTaskPnt > 1) {ServoTaskPnt = -1;}  // centre servo
            break;
          case 1: // in 'Tasks' display mode, so toggle tasks
            DispPnt++; if (DispPnt > 4) {DispPnt = 0;}
            DispMode = 0; // reset the display mode
            break;
          case 2: // in 'Sumo' display mode, so toggle modes
            DispPnt++; if (DispPnt > 0) {DispPnt = 0;}
            DispMode = 0; // reset the display mode
            break;
        }
      } else {
        // extended press so decrement the display backwards
        LED_Task = -1;    // set LED tasks to default unless changed here
        DispChng = true;  // flag to initialise certain functions
        if (DispMenu == 0) {
          DispPnt--; if (DispPnt < 0) {DispPnt = 4;}
          if (DispPnt == 4) { // turn ON fast read mode
            ADC_Fast = true;
              initialiseIR(); LED_Task = -20;
          } else {
            ADC_Fast = false; LED_Task = -1;
          }
        }
        if (DispMenu == 1) {DispPnt--; if (DispPnt < 0) {DispPnt = 4;}}
        if (DispMenu == 2) {DispPnt--; if (DispPnt < 0) {DispPnt = 0;}}
      }
      sw1Cnt = 0; sw1Timer = 0; sw1DwnTime= 0; DispCnt = 0;
    }
  }
  sw1LastState = sw1State; // record current state
  if (sw1Cnt > 0) {sw1Timer++;}
}

// ---------------------------------------------------------------------------------

void readWiFiRx() {
  // called from the main loop every cycle to check for ESP-NOW data  
  // a state machine is used to separate the data blocks by header type
  DispDel = 500;  // hold off display updates for 10 sec
  switch(Rx_Task) {
    case 0: // determine the type of data to handle
           if (Rx_Buff.ESPdata[Rx_Pnt] == 'B') {Rx_Task = 3; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'C') {WiiType = 'C'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == 'N') {WiiType = 'N'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '-') {WiiType = '-'; Rx_Task = 1; break;}
      else if (Rx_Buff.ESPdata[Rx_Pnt] == '$') {Rx_Task = 2; break;}
      // oh dear! bloick header not recognised so discard the block and flag error
      else {WiFiRxErr++; WiFiRx = false;}
      break;
      
    case 1: // load Wii I2C data
      // after the header byte there are 6 data bytes + checksum
      // perform XOR on 6 bytes to check for errors
      if (Rx_Buff.ESPdata[7] == (Rx_Buff.ESPdata[1] ^ Rx_Buff.ESPdata[2] ^ Rx_Buff.ESPdata[3] ^ Rx_Buff.ESPdata[4] ^ Rx_Buff.ESPdata[5] ^ Rx_Buff.ESPdata[6])) {
        // data has passed the checksum test
        RxWiFi[0] = Rx_Buff.ESPdata[1];
        RxWiFi[1] = Rx_Buff.ESPdata[2];
        RxWiFi[2] = Rx_Buff.ESPdata[3];
        RxWiFi[3] = Rx_Buff.ESPdata[4];
        RxWiFi[4] = Rx_Buff.ESPdata[5];
        RxWiFi[5] = Rx_Buff.ESPdata[6];

        // if Nunchuk data strip off the JX,JY and CZ values
        RxRec = true; // indicate a valid frame
        if (WiiType == 'N') {
          JoyX = RxWiFi[0];   // joystick X 0-255
          JoyY = RxWiFi[1];   // joystick Y 0-255
          CZ = RxWiFi[5] & 3; // CZ buttons, C = 2/0, Z = 1/0, active low
        } else {
          // not Nunchuk so set defaults
          JoyX = 128; JoyY = 128; CZ = 3;
        }
      } else {
        WiFiRxErr++;
      }
      Rx_Task = 0;  // reset the task pointer to lok at another block if needed
      Rx_Pnt += 8;  // increase the data pointer beyond this block
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0;} // reached end of buffer
      break;

    case 2: // load text data as if from a serial port
      // read characters up to the end of the buffer
      Rx_Pnt++;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; Rx_len = 0; break;} // reached end of buffer
      keyVal = Rx_Buff.ESPdata[Rx_Pnt]; PrintTgt = 1; decodeKey(keyVal);
      Serial.write(keyVal);
      break;

    case 3: // load binary data block
      // data length is specified in second byte
      Rx_Pnt++;
      // at the moment just jump over it, binary trasnfer is not needed
      Rx_Pnt += Rx_Buff.ESPdata[Rx_Pnt] + 1;
      if (Rx_Pnt >= Rx_len) {WiFiRx = false; break;} // reached end of buffer
      break;
  }
}

// ---------------------------------------------------------------------------------

void readWii() {
  // called every 20ms, but wireless transceiver only transmits every 40ms
  // considers data received from the Wii Nunchuk controller over WiFi and 
  // manages all movements demanded by the controller
  if (readWiiCall) {return;}
  readWiiCall = true; // prevent re-entry of this function
  
  if (RxRec) {
  //##############################################################################
  //
  //  WiFi Nunchuk data received
  //
  //##############################################################################
    RxRec = false; overRun = 5;
//    PrintTx += "JX" + String(JoyX) + " JY" + String(JoyY) + " CZ" + String(CZ) + "\n";
//    Serial.print("CZ "); Serial.println(CZ);
    if (!WiFiEn) {
  //##############################################################################
  //
  //  WiFi is currently disabled, listening
  //
  //##############################################################################
      // we have not received enough 'C' buttons to make the system active
      // the 'C' button alone must be held in to activate the Wii WiFi link
      if (CZ == 1) {
        // C button only is being pressed
        WiFiCntC += 4;
//        Serial.println("WiFiCnt = " + String(WiFiCntC));
        if (WiFiCntC > 60) {
          // going active, user has held in 'C' button for sufficient time
          Serial.println("WiFiEn = true");
          WiFiCntC = 64; WiFiEn = true; C_Dn = 8;
          LED_Task = -1;  // reset LED tasks
//          if (MainTask != 0) {SetMainTask(0);}
//          if (!AtRest) {SetMainTask(98); LEDMthTask = -1;} // move to squat (default)
          // set defaults here
//          Gear = 1;
        } else {
          // still counting C buttons
          LEDVal = WiFiCntC/10; LED_Task = -10;
        }
      } else {
        // C button not pressed or released early so check for a inadequate count
        if (WiFiCntC > 0) {
          // reduce the C-button counter, which has to be sustained to enable WiFi mode
          WiFiCntC--; LEDVal = 1 + (WiFiCntC/10);
          if (WiFiCntC < 1) {LED_Task = -1;} else {LED_Task = -10;}
        }
      }
    } else {
      
  //##############################################################################
  //
  //  WiFi is enabled
  //
  //##############################################################################
      // WiFi enabled so check for power down
      if (CZ == 0) {
        // both C and Z buttons are being pressed so move towards powering down
        if (WiFiCntC > 0) {
          WiFiCntC -= 5;
  //        Serial.println("WiFiCnt = " + String(WiFiCntC));
          if (WiFiCntC <= 0) {
            // power-down condition satisfied, so go to rest
            Serial.println("WiFi disabled!");
            WiFiCntC = 0; disableWiFi();
            // set defaults here
             LED_Task = -1;
          } else {
            // still counting down CZ buttons
            LEDVal = WiFiCntC/10; LED_Task = -10;
          }
        }
      } else {
        // Check CZ buttons to see if a 'Gear' change is wanted
        if ((CZ & 2) == 0) {
          // C button is pressed, assume gear change
          C_Dn++; if (C_Dn == 2) {
//            C_Dn = 8; if (Gear < 5) {Gear++;}
//            if (LEDTaskPnt != 99) {LEDNextTask = LEDTaskPnt;}
//            LEDTaskPnt = -5; // display 'Gear' colour in the eyes
          }
//          Serial.print("Gear "); Serial.println(Gear);
        } else {C_Dn = 0;} // C button released
        // Check Z button. It has two functions. Short press change gear, but if
        // held in it changes the mode of the joysticks
        if ((CZ & 1) == 0) {
          // Z button is pressed
          Z_Dn++; if (Z_Dn > 12) {Z_Dn = 50;} // Z_Dn capped at 50, held for >= 2 seconds
//          Serial.print("Z_Dn "); Serial.println(Z_Dn);
        } else {
          if ((Z_Dn > 0) && (Z_Dn <= 10)) {
            // held down for less than 400 ms @ 40ms Rx cycle
//            if (Gear > 1) {Gear--;}  // short press so change gear if possible
//            if (LEDTaskPnt != 99) {LEDNextTask = LEDTaskPnt;}
//            LEDTaskPnt = -5; // display 'Gear' colour in the eyes
//            Serial.print("Gear "); Serial.println(Gear);
          } Z_Dn = 0; // Z button released
        }
//        Serial.print("Z_Dn "); Serial.println(Z_Dn);

        if (WiFiCntC < 60) {
          // C-button counter is not at full strength, assume WiFi off released prematurely
          WiFiCntC++; LEDVal = 1 + (WiFiCntC/10);
          if (WiFiCntC == 60) {LED_Task = -1;} else {LED_Task = -10;}
        }
      }

      
  //##############################################################################
  //
  //  Wii joystick demands
  //
  //##############################################################################
      // now respond to joystick demands. We use a state machine to control how
      // the joystick demands are responded to. But first we resolve the JoyX and
      // JoyY values into a strength vector JoyV, and a direction JoyD
      // when not centred a counter is used to record how long the joystick is off centre
//      Serial.print("X"); Serial.print(JoyX); Serial.print("  Y"); Serial.println(JoyY);
      // During Sumo AI count down the joystick demand is inhibited until the count
      // reaches zero
      if (!JoyEn) {JoyX = 128; JoyY = 128;} // inhibit joystick demands during Sumo count-down

      JoyXC = JoyX - 128; JoyYC = JoyY - 128;
      if ((abs(JoyXC) < 8) && (abs(JoyYC) < 8)) {
        // joystick is in centre deadband region
        JoyV = 0; JoyD = 0; // clear vector and direction values 
      } else {
        // calculate vector outside of centre region
        JoyV = int(sqrt(sq(float(JoyXC)) + sq(float(JoyYC))));
        JoyV = min(int(JoyV),127);    // cap max value, which can reach 181
        if (JoyXC == 0) {JoyXC = 1;}  // prevent divide by zero
        if (JoyYC == 0) {JoyYC = 1;}
        
        JoyD = 0; // reset direction pointer to default
        if (JoyYC >= 8) {
          // in upper hemisphere
          if (abs(JoyYC/JoyXC) >= 2) {JoyD = 1;}          // pointing N
          else if (abs(JoyXC/JoyYC) >= 2) {
            if (JoyXC > 0) {JoyD = 3;}                    // pointing E
            else {JoyD = 7;}                              // pointing W
          } else {
            if (JoyXC > 0) {JoyD = 2;}                    // pointing NE
            else {JoyD = 8;}                              // pointing NW
          }
        } else if (JoyYC > -8) {
          // in middle area
          if (JoyXC >= 8) {JoyD = 3;}                    // pointing E
          else {JoyD = 7;}                                // Pointing W
        } else {
          // in lower hemisphere
          if (abs(JoyYC/JoyXC) >= 2) {JoyD = 5;}          // pointing S
          else if (abs(JoyXC/JoyYC) >= 2) {
            if (JoyXC > 0) {JoyD = 3;}                    // pointing E
            else {JoyD = 7;}                              // pointing W
          } else {
            if (JoyXC > 0) {JoyD = 4;}                    // pointing SE
            else {JoyD = 6;}                              // pointing SW
          }
        }
      }
//      Serial.print("JoyV "); Serial.print(JoyV);
//      Serial.print("   JoyD "); Serial.println(JoyD);
//      PrintTx += String(JoyD) + "\t" + String(JoyV) + "\n"; JoyD = 0; // print joystick vectors but don't move
//      if (MoveState > 0) {PrintTx += String(Speed) + "\n";}
      
      
  //##############################################################################
  //
  //  Respond to Wii controller
  //
  //##############################################################################
//      Serial.println("JoyD " + String(JoyD) + "  MoveState " + String(JoyD));
      switch(JoyD) {
        case 0:
          // default state, SumoBot will be in ready state or rest state
          if (JoyDL != 0) {
            // look for 'Z' button, if pressed break harshly
            if ((CZ & 1) == 0) {MoveSTOP(1);} else {MoveSTOP(0);}
          } break;
        case 1: // moving forward
          MoveFwd(); break;
        case 2: // moving forward bearing to the right
          MoveFwdRht(); break;
        case 3: // neutral turning right
          MoveTurnRight(); break;
        case 4: // moving backward bearing to the right
          MoveBkdRight(); break;
        case 5: // moving backward
          MoveBkd(); break;
        case 6: // moving backward bearing to the left
          MoveBkdLeft(); break;
        case 7: // neutral turning left
          MoveTurnLeft(); break;
        case 8: // moving forward bearing to the left
          MoveFwdLft(); break;
      }
      // track joystick demands for centred position
      JoyDL = JoyD;
    }
  } else {
  //##############################################################################
  //
  //  No WiFi Nunchuk received
  //
  //##############################################################################
    if (WiFiEn) {
      // droid is enabled for Wii WiFi control
      // maintain Wii enabled counter
      if ((WiFiCntC < 64) && (CZ != 0)) {
        WiFiCntC++; LEDVal = WiFiCntC/10;
        if (WiFiCntC == 64) {LED_Task = -1;} else {LED_Task = -10;}
      }
      
      // if in MoveState == 20 complete the move first, otherwise do these checks
      if (MoveState != 20) {
        // if moving ensure that demands are still being received, if not goto ready
        if (MoveState != 0) {
          overRun --;
//          if (overRun < 1) {MoveState = 0; MoveToReady(1); loopWhileWalk(); LEDMthTask = -1;}
        }
        // perform auto-rest function if no demands
        if (WiFiRestCnt > 0) {WiFiRestCnt--;}
        if (WiFiRestCnt == 1) {
          // if not at rest go there
//          if (!AtRest) {MoveToStandFromAny();}
        }
      }
//      Serial.println(WiFiRestCnt);
    }
  }

  if (WiFiEn) {
  //##############################################################################
  //
  //  WiFi background tasks when WiFi is enabled
  //
  //##############################################################################
    // only do these functions if WiFi is enabled
    if (MoveState == 20) {
      // long un-demanded moves like bowing are watched until they complete
      MoveState = 0; LED_Task = -1;
    }
    
    if (MoveState == -1) {
      // at the end of a joystick move we want to set things back to normal
      LED_Task = -1;
//      SetSpeedDefaults();
      MoveState = 0;
    }
  }
  
  readWiiCall = false; // re-allow calls to this function
//  Serial.print("CntZ "); Serial.println(WiFiCntZ);
//   Serial.println(MoveState);
}

// --------------------------------------------------------------------------------

void reportCalData() {
  // preint out the calibration values
  Serial.println("");
  Serial.println("AvBlk[" + String(AvBlk[0]) + "," + String(AvBlk[1]) + "," + String(AvBlk[2]) + "," + String(AvBlk[3]) + "," + String(AvBlk[4]) + "]");
  Serial.println("AvWht[" + String(AvWht[0]) + "," + String(AvWht[1]) + "," + String(AvWht[2]) + "," + String(AvWht[3]) + "," + String(AvWht[4]) + "]");
  Serial.println("AcXOff " + String(AcXOff) + "\tAcYOff " + String(AcYOff) + "\tAcZOff " + String(AcZOff));
  Serial.println("GyXOff " + String(GyXOff) + "\tGyYOff " + String(GyYOff) + "\tGyZOff " + String(GyZOff));
  Serial.println("ServoMin " + String(ServoMin) + "\tServoCtr " + String(ServoCtr) + "\tServoMax " + String(ServoMax));
}

// --------------------------------------------------------------------------------

void RxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Rx buffer of length zlen
  Rx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    Rx_Chk = Rx_Chk ^ Rx_Buff.ESPdata[zP];
  }
}

// ---------------------------------------------------------------------------------

void ServoOFF() {
  // detaches the servo to stop it consuming power
  microServo.detach(); ServoEn = false;
}

// ---------------------------------------------------------------------------------

void ServoON() {
  // attaches the servo and sets its value
  // we do not need to set the min/max values; 544/2400 as we write PWM directly
  microServo.attach(servoPin); // assign pin
  microServo.writeMicroseconds(servoVal); ServoEn = true;
}

// ---------------------------------------------------------------------------------

void Servo_Task() {
  // servo movement tasks, called every 20ms, or when LTOF_On == 5
  // -ve initiation values are:
  // -1   return to centre position, pwr off
  // -2   scan for a target
  if (ServoTaskPnt == 0) {return;}
  if (ServoTaskDel > 0) {ServoTaskDel--; return;} // servo task delay skip timer

  bool zReport = false; // when true report data to serial port
  int16_t zDP = 0;      // data store pointer
  int16_t zP;           // For... loop pointer

  if (ServoTaskPnt < 0) {
    // set -ve to initiate variables
    ServoTaskPnt = abs(ServoTaskPnt); ServoSubTask = 0;
  }
  switch(ServoTaskPnt) {
    case 1: // return servo to centre position, allow time, then PWM off
      switch(ServoSubTask) {
        case 0:
          servoVal = ServoCtr; ServoON(); ServoTaskCnt = 50;
          ServoSubTask++; break;
        case 1:
          ServoTaskCnt--; if (ServoTaskCnt < 1) {
            ServoOFF(); ServoTaskPnt = 0;
          } break;
      } break;

    case 2: // scan for a target, storing data and key values
      // case 0 - initialise subtasks
      // case 1 - sweep scan to the right, angle increasing
      // case 2 - sweep scan to the left, angle decreasing
      // case 3 - stop sweep scan and centralise
      // case 4 - wait for edge turning to end, then recommence scan
      switch(ServoSubTask) {
        case 0:
          // initialise scanning process
          ServoON(); clearRangeData();
          if (ServoAng <= 90) {ServoSubTask = 1;} else {ServoSubTask = 2;}
          ServoAngMin = ServoAngMinDef; ServoAngMax = ServoAngMaxDef;
          ServoAngTgtLL = ServoAngMin; ServoAngTgtUL = ServoAngMax;
          RangeFtd = RangeMax; RangeFtdMin = RangeMax; ServoTgtRec= false;
          RangeFtdTgt = RangeMax; TgtOR = true;
          RangeFtdTgt10 = RangeMax + 10; ServoTgt10 = false;
          ServoAngTgtAvg = 90;  // assumes target is central to start with
          ServoInc = 1; if (ServoAuto) {ServoInc = 10;}
          break;
          
        case 1:
          // sweep with increasing angle towards ServoAngMax
          // the minimum range point angle is stored in ServoAngTmp
          zDP = ServoAng - ServoAngMinDef;
          RangeData[zDP] = RangeFtd; // store range before changing angle
          if (ServoInc > 1) {
            // fill in the missing data
            for (zP = 1; zP < ServoInc; zP++) {
              zDP++; if (zDP <= ServoAngWidth) {RangeData[zDP] = RangeFtd;}
            }
          }
          // look for new minimum value, if so record range and angle
          if (RangeFtd < RangeFtdMin) {
            RangeFtdMin = RangeFtd; ServoAngTmp = ServoAng;
            RangeFtdTgt10 = RangeFtdMin + 10; // lower the cut-off point limit
            ServoTgt10 = false;
          } else if (RangeFtd >= RangeFtdTgt10) {
            // record the 10mm cut-off angle if detected
            if (!ServoTgt10) {
              ServoAngTgt10 = ServoAng; ServoTgt10 = true;
//              Serial.print(String(ServoSubTask) + "\tUL: " + String(ServoAng) + "\n");
            }
          }
          ServoAng+= ServoInc;  // change the angle
          // look for max angle, and remove any over-shoot
          if (ServoAng > ServoAngMax) {ServoAng = ServoAngMax;}
          setServoAng(ServoAng);  // set the servo angle
//          Serial.print(String(ServoSubTask) + "\t" + String(ServoAng) + "\t" + String(RangeFtd) + "\n");
          if (ServoAng == ServoAngMax) {
            // max angle limit reached, so change direction
            ServoSubTask = 2;
            // if we did not see the cut-off set UL angle to max
            if (!ServoTgt10) {ServoAngTgt10 = ServoAngMax;}
            ServoAngTgtUL = ServoAngTgt10;
            ServoTgt10 = false;             // reset the cut-off flag
            // report values on direction change
            zReport = true;
            RangeFtdTgt = RangeFtdMin; RangeFtdMin = RangeMax;
            ServoAngTgtAvg = ServoAngTgt + ((ServoAngTmp - ServoAngTgt)/2);
            ServoAngTgt = ServoAngTmp;
            RangeFtdTgt10 = RangeMax + 10;  // reset cut-off range
            if (ServoAuto) {
              // if not at full sweep, clear top data
              zDP++;
              while (zDP <= ServoAngWidth) {
                RangeData[zDP] = RangeMax; zDP++;
              }
              // in auto mode scan angle steps are increased
              if (RangeFtdTgt >= 500){
                // long range so use full scan range
                ServoAngMin = ServoAngMinDef; TgtOR = true;
              } else {
                // reduce scan range as target range gets closer
                if (TgtOR) {ServoAngTgtAvg = ServoAngTgt;} // 1st in-range value
                ServoAngTmp = map(RangeFtdTgt,50,500,10,40);
                ServoAngMin = ServoAngTgtAvg - ServoAngTmp;
                if (ServoAngMin < ServoAngMinDef) {ServoAngMin = ServoAngMinDef;}
                TgtOR = false;  // target within range
              }
              ServoInc = (ServoAngMax - ServoAngMin)/8;
            } else {
              ServoInc = 1; ServoAngMin = ServoAngMinDef;
            }
            // calculate target centre based on cut-off limits
//            ServoAngTgt = ServoAngTgtLL + ((ServoAngTgtUL - ServoAngTgtLL)/2);
            ServoTgtRec= true;
          }
          if (ServoAuto) {ServoTaskDel = 1;}  // larger steps but with delays
          break;

        case 2:
          // sweep with decreasing angle towards ServoAngMin
          // the minimum range point angle is stored in ServoAngTmp
          zDP = ServoAng - ServoAngMinDef;
          RangeData[zDP] = RangeFtd; // store range before changing angle
          if (ServoInc > 1) {
            for (zP = 1; zP < ServoInc; zP++) {
              zDP++; if (zDP <= ServoAngWidth) {RangeData[zDP] = RangeFtd;}
            }
          }
          // look for minimum value, if so store range and angle
          if (RangeFtd < RangeFtdMin) {
            RangeFtdMin = RangeFtd; ServoAngTmp = ServoAng;
            RangeFtdTgt10 = RangeFtdMin + 10; // adjust the cut-off point limit
            ServoTgt10 = false;
          } else if (RangeFtd >= RangeFtdTgt10) {
            // record the 10mm cut-off angle if detected
            if (!ServoTgt10) {
              ServoAngTgt10 = ServoAng; ServoTgt10 = true;
//              Serial.print(String(ServoSubTask) + "\tLL: " + String(ServoAng) + "\n");
            }
          }
          ServoAng-= ServoInc;  // change the angle
          // look for max angle, and remove any over-shoot
          if (ServoAng < ServoAngMin) {ServoAng = ServoAngMin;}
          setServoAng(ServoAng);  // set the servo angle
//          Serial.print(String(ServoSubTask) + "\t" + String(ServoAng) + "\t" + String(RangeFtd) + "\n");
          if (ServoAng == ServoAngMin) {
            // min angle limit reached, so change direction
            ServoSubTask = 1;
            // if we did not see the cut-off set UL angle to max
            if (!ServoTgt10) {ServoAngTgt10 = ServoAngMin;}
            ServoAngTgtLL = ServoAngTgt10;
            ServoTgt10 = false;             // reset the cut-off flag
            // report values on direction change
            zReport = true;
            RangeFtdTgt = RangeFtdMin; RangeFtdMin = RangeMax;
            ServoAngTgtAvg = ServoAngTgt + ((ServoAngTmp - ServoAngTgt)/2);
            ServoAngTgt = ServoAngTmp; 
            RangeFtdTgt10 = RangeMax + 10;  // reset cut-off range
            if (ServoAuto) {
              // if not at full sweep, clear bottom data
              zDP-= ServoInc;
              while (zDP >= 0) {
                RangeData[zDP] = RangeMax; zDP--;
              }
              // in auto mode scan angle steps are increased
              if (RangeFtdTgt >= 500){
                // long range so use full scan range
                ServoAngMax = ServoAngMaxDef; TgtOR = true;
              } else {
                // reduce scan range as target range gets closer
                if (TgtOR) {ServoAngTgtAvg = ServoAngTgt;} // 1st in-range value
                ServoAngTmp = map(RangeFtdTgt,50,500,10,40);
                ServoAngMax = ServoAngTgtAvg + ServoAngTmp;
                if (ServoAngMax > ServoAngMaxDef) {ServoAngMax = ServoAngMaxDef;}
                TgtOR = false;  // target within range
              }
              ServoInc = (ServoAngMax - ServoAngMin)/8;
            } else {
              ServoInc = 1; ServoAngMax = ServoAngMaxDef;
            }
            // calculate target centre based on cut-off limits
//            ServoAngTgt = ServoAngTgtLL + ((ServoAngTgtUL - ServoAngTgtLL)/2);
            ServoTgtRec= true;
          } 
          if (ServoAuto) {ServoTaskDel = 1;}  // larger steps but with delays
          break;

        case 3:
          // edge survival mode, so centralise LTOF servo
          ServoAng = 90; setServoAng(ServoAng);  // set the servo angle
          ServoAngMin = ServoAng - 10; ServoAngMax = ServoAng + 10;
          clearRangeData();
          ServoAngTgtLL = ServoAngMin; ServoAngTgtUL = ServoAngMax;
          RangeFtd = RangeMax; RangeFtdMin = RangeMax; ServoTgtRec= false;
          RangeFtdTgt = RangeMax; TgtOR = true;
          RangeFtdTgt10 = RangeMax + 10; ServoTgt10 = false;
          ServoAngTgtAvg = 90;  // assumes target is central to start with 90 Deg.
          ServoSubTask++; break;

        case 4:
          // wait for edge turning to end, whilst monitoring range value
          RangeFtdTgt = RangeFtd;
          if ((Turn_Flg == 0) && (RndTurn_Flg == 0)) {
            // time to go, edge turning and/or random turning has stopped
            RangeFtdTgt = RangeMax;
            ServoTgt10 = false;             // reset the cut-off flag
            if (Turning > 0) {ServoSubTask = 1;}  // turning right so scan towards max
            else {ServoSubTask = 2;}        // turning left so scan towards min
          } break;
      } break;
  }
//  if (zReport) {
//    Serial.print(String(ServoSubTask) + "\t");  // task pointer
//    Serial.print(String(ServoAngTgtLL) + "\t"); // lower cut-off angle
//    Serial.print(String(ServoAngTgt) + "\t");   // target angle
//    Serial.print(String(ServoAngTgtUL) + "\t"); // upper cut-off angle
//    Serial.print(String(RangeFtdTgt) + "\n");   // target range
//  }
}

// ---------------------------------------------------------------------------------

void SetLX16Value() {
  // sends a speed value to the target LX-16A servo
  LX16Spd = cmdVal;
  LobotSerialServoSetMode(Serial2, LX16Tgt, 1, LX16Spd);  
  PrintTx += "LV=" + String(cmdVal) + "\n";
}

// ---------------------------------------------------------------------------------

void setMainTask(int zMT) {
  // change the main task pointer
  // set default conditions for starting/ending a task
  ADC_Fast = false;       // slow down sensor scanning
  AI_ATTACK_EN = false;   // if == true it enables AI attack mode
  AI_EDGE_ON = false;     // if == true AI performs edge detection
  AI_IR_ON = false;       // if == true AI turns ON IR sensor scanning
  AI_LED_Task = -20;      // default LED task when in IR mode
  AI_LTOF_ON = false;     // == true enables LTOF servo scanner in AI mode
  AI_Sub = 0;             // initial AI sub-task
  AI_TIP_ON = false;      // == true enables AI tip responses
  AI_Title$ = "";         // title used in AI display logo
  CountDown = 0;          // seconds count down used by AI
  DispMode = 0;           // set display to default option
  DispNOP = false;        // release display lock
  DistanceMs = millis();  // reset the distance travelled timer
  ItsOver = false;        // == true if the bot has fallen over
  JoyEn = true;           // nunchuk joystick enable flag, default == true
  L_Dst = 0;              // distance travelled in 0.01mm increments
  LED_Del = 0;            // reset LED task pause delay
  LED_IR_Mask = 0;        // used to block unnecessary updates
  LED_Task = -1;          // default LED task
  MainCnt = 0;            // Main Task general counter/timer
  MainDel = 0;            // Main task delay counter, if > 0 holds off tasks
  MainNext = 0;           // next MainTask to be performed after a delay
  MainSub = 0;            // Main Task sub-task
  MoveTask = 0;           // special move task pointer
  RangeFtdTgt = RangeMax; // prevent false ranging
  R_Dst = 0;              // distance travelled in 0.01mm increments
  RndAhead = 0;           // if > 0 then go straight ahead for RndAhead mm
  ServoAuto = false;      // switch OFF servo auto-adjustment features
  ServoInc = 1;           // reset angular increment
  ServoOFF();             // when changing tasks turn servo off
  ServoTaskDel = 0;       // servo task delay skip timer
  ServoTaskPnt = -1;      // centre servo scanner
  setWheelSpeed(0,0);     // stop wheels and set L_Spd & R_Spd to zero
  SpeedInc = 4;           // auto speed adjust increment, default is 4
  SpeedSrt = 0;           // starting speed in AI mode
  Steer = 0;              // turning factor, +ve right, -ve left
  TipLast = 0;            // previous Tip value
  Turn_Flg = 0;           // clear turn flag
  VL53L0X_OFF();          // turn LTOF ranger OFF
  
  MainTask = zMT;
}

// ---------------------------------------------------------------------------------

void setServoAng(int zSA) {
  // called to move the servo to a specific angle
  servoVal = map(zSA,50,130,ServoMax,ServoMin);
  if (!ServoEn) {microServo.attach(servoPin);}  // servo was not attached
  microServo.writeMicroseconds(servoVal); ServoEn = true;
}

// ---------------------------------------------------------------------------------

void shiftBatData() {
  // shift battery data so that it is a rolling average
  for (int zB = 0; zB < 59; zB++) {BatData[zB] = BatData[zB + 1];}
}

// ---------------------------------------------------------------------------------

void VL53L0X_OFF() {
  // disable the VL53L0X for reading ranges
  digitalWrite(XSHUT,LOW);  // VL53L0X reset is active LOW
  LTOF_On = 0;              // > 0 when laser ranging is enabled, default is 0 == OFF
}

// ---------------------------------------------------------------------------------

void VL53L0X_ON() {
  // enable the VL53L0X for reading ranges
  digitalWrite(XSHUT,HIGH); // enable the VL53L0X
  VL53L0X_init(true); delay(10);
  VL53L0X_setTimeout(500);

  // Start continuous back-to-back mode (take readings as
  // fast as possible).  To use continuous timed mode
  // instead, provide a desired inter-measurement period in
  // ms (e.g. sensor.startContinuous(100)).
  VL53L0X_startContinuous(); delay(10);
  LTOF_On = 1; // restart the measurement statemachine
}

// --------------------------------------------------------------------------------

void waitMainEnd() {
  // called by either switch function when a button is pressed during a task
//  Serial.println("waitMainEnd");
  setMainTask(0); waitWhileDwn();
}

// --------------------------------------------------------------------------------

void waitWhileDwn() {
  // called by either switch function when a button is pressed during a task
  MoveSTOP(1);  // stop any movement immediately
  Any = 0;      // zero loop counter
  while ((digitalRead(sw0Pin) == LOW) || (digitalRead(sw1Pin) == LOW) || !DMsw0 || !DMsw1) {
    // do nothing while either button is pressed
    delay(1); Any++; yield();
    if (DispMon) {readSerial();} // read serial port if connected to Display Monitor
    if (Any >= 200) {
      // still test for long press RESET every 200ms
      if (digitalRead(sw1Pin) == LOW) {
        ResetCnt++; if (ResetCnt >= 15) {
          // button held down for >= 3 seconds so force a soft reset
          Display_Reboot(); delay(1000);
          resetFunc();        //call reset
        }
      } else {ResetCnt = 0;}  // clear the timer
      Any = 0;                // reset loop counter
    }
  }
  // reset switch variables
  sw0Cnt = 0; sw0Timer = 0; sw0LastState = HIGH; ESC0 = false; SW0_Nop = false;
  sw1Cnt = 0; sw1Timer = 0; sw1LastState = HIGH; ESC1 = false; SW1_Nop = false;
  DMsw0 = true; DMsw1 = true;
  synchLoopTimers(0);
}

// --------------------------------------------------------------------------------




  //###############################################################################
  //
  //  WiFi Code
  //
  //###############################################################################

void WiFiClearRxBuff(int zNum) {
  // clears the contents of the receive buffer
  for (int zP = 0; zP < zNum;zP++) {Rx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiClearTxBuff(int zNum) {
  // clears the contents of the transmit buffer
  for (int zP = 0; zP < zNum;zP++) {Tx_Buff.ESPdata[zP] = 0;}
}

// --------------------------------------------------------------------------------

void WiFiDisconnect() {
  // enter the WiFi disconnected state
  WiFiConnected = false;
  WiFiClearRxBuff(WiFiTx_len);
  WiFiClearTxBuff(WiFiTx_len);
  WiFiTryCnt = 0; // reset the connection retry counter
  WiFiTx_CB = 1;  // set ready to send flag
//  Serial.println("WiFiDisconnected!");
}

// --------------------------------------------------------------------------------

void WiFiTryToConnect() {
  // try to connect with the Wii Transceiver by sending devices name
  // initialise ESP-NOW link first
  Init_ESP_NOW();
  
  Tx_Buff.ESPdata[0] = 'S';
  Tx_Buff.ESPdata[1] = 'u';
  Tx_Buff.ESPdata[2] = 'm';
  Tx_Buff.ESPdata[3] = 'o';
  Tx_Buff.ESPdata[4] = 'B';
  Tx_Buff.ESPdata[5] = 'o';
  Tx_Buff.ESPdata[6] = 't';
  WiFiTx_len = 7;
//  t0 = micros();
  esp_err_t result = esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
  if (result == ESP_OK) {
//    Serial.println("Sent with success");
  } else {
//    Serial.println("Error sending the data");
  } 

  WiFiTryNum++; // count the total number of attempts
//  Serial.println("WiFiTryToConnect...");
}

// --------------------------------------------------------------------------------

void WiFiTxGetChecksum(int zlen) {
  // performs an XOR checksum calc on the Tx buffer of length zlen
  WiFiTx_Chk = 0x55;  // define checksum seed as 01010101
  for (int zP = 0;zP < zlen;zP++) {
    WiFiTx_Chk = WiFiTx_Chk ^ Tx_Buff.ESPdata[zP];
  }
}

// ---------------------------------------------------------------------------------
